/* GStreamer
 * Copyright (C) 1999,2000 Erik Walthinsen <omega@cse.ogi.edu>
 *                    2000 Wim Taymans <wim.taymans@chello.be>
 *
 * gstcomedisink.c: 
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <unistd.h>
#include <errno.h>

#include <gstcomedisink.h>
#include <gstcomediclock.h>

static GstElementDetails gst_comedisink_details = {  
	longname: "Audio Sink (COMEDI)",
	klass: "Sink/Audio",
	description: "Output to a sound card via COMEDI",
	version: VERSION,
	author: "Erik Walthinsen <omega@cse.ogi.edu>, "
		"Wim Taymans <wim.taymans@chello.be>",
	copyright: "(C) 1999",
};

static void 			gst_comedisink_class_init		(GstComediSinkClass *klass);
static void 			gst_comedisink_init		(GstComediSink *comedisink);
static void 			gst_comedisink_finalize		(GObject *object);

static gboolean 		gst_comedisink_open_audio		(GstComediSink *sink);
static void 			gst_comedisink_close_audio		(GstComediSink *sink);
static gboolean 		gst_comedisink_sync_parms 		(GstComediSink *comedisink);
static GstElementStateReturn 	gst_comedisink_change_state	(GstElement *element);

static GstPadConnectReturn	gst_comedisink_sinkconnect		(GstPad *pad, GstCaps *caps);

static void 			gst_comedisink_set_property	(GObject *object, guint prop_id, const GValue *value, 
		  						 GParamSpec *pspec);
static void 			gst_comedisink_get_property	(GObject *object, guint prop_id, GValue *value, 
								 GParamSpec *pspec);

static void 			gst_comedisink_chain		(GstPad *pad,GstBuffer *buf);

/* ComediSink signals and args */
enum {
  SIGNAL_HANDOFF,
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_DEVICE,
  ARG_MUTE,
  ARG_FORMAT,
  ARG_CHANNELS,
  ARG_FREQUENCY,
  ARG_FRAGMENT,
  ARG_BUFFER_SIZE
  /* FILL ME */
};

GST_PAD_TEMPLATE_FACTORY (comedisink_sink_factory,
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW (
    "comedisink_sink",
    "x-gst-sci/time-series",
      "format",     GST_PROPS_LIST (
         	      GST_PROPS_STRING("uint16"),
         	      GST_PROPS_STRING("float32"),
         	      GST_PROPS_STRING("float64")
      		    ),
      "rate",       GST_PROPS_INT_RANGE (1, 1000000000),
      "channels",   GST_PROPS_INT_RANGE (1, 16)
  )
);

#define GST_TYPE_COMEDISINK_CHANNELS (gst_comedisink_channels_get_type())
static GType 
gst_comedisink_channels_get_type(void) {
  static GType comedisink_channels_type = 0;
  static GEnumValue comedisink_channels[] = {
    {0, "0", "Silence"},
    {1, "1", "Mono"},
    {2, "2", "Stereo"},
    {0, NULL, NULL},
  };
  if (!comedisink_channels_type) {
    comedisink_channels_type = g_enum_register_static("GstAudiosinkChannels", comedisink_channels);
  }
  return comedisink_channels_type;
}


static GstElementClass *parent_class = NULL;
static guint gst_comedisink_signals[LAST_SIGNAL] = { 0 };

GType
gst_comedisink_get_type (void) 
{
  static GType comedisink_type = 0;

  if (!comedisink_type) {
    static const GTypeInfo comedisink_info = {
      sizeof(GstComediSinkClass),
      NULL,
      NULL,
      (GClassInitFunc)gst_comedisink_class_init,
      NULL,
      NULL,
      sizeof(GstComediSink),
      0,
      (GInstanceInitFunc)gst_comedisink_init,
    };
    comedisink_type = g_type_register_static (GST_TYPE_ELEMENT, "GstComediSink", &comedisink_info, 0);
  }

  return comedisink_type;
}

static GstBufferPool*
gst_comedisink_get_bufferpool (GstPad *pad)
{
  GstComediSink *comedi;
  
  comedi = GST_COMEDISINK (gst_pad_get_parent(pad));

  return comedi->sinkpool;
}

static void
gst_comedisink_finalize (GObject *object)
{
  GstComediSink *comedisink = (GstComediSink *) object;

  g_free (comedisink->device);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
gst_comedisink_class_init (GstComediSinkClass *klass) 
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = g_type_class_ref(GST_TYPE_ELEMENT);

  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_DEVICE,
    g_param_spec_string("device","device","device",
                        "/dev/dsp",G_PARAM_READWRITE)); /* CHECKME! */
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_MUTE,
    g_param_spec_boolean("mute","mute","mute",
                         TRUE,G_PARAM_READWRITE)); 

  /* it would be nice to show format in symbolic form, oh well */
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_FORMAT,
    g_param_spec_int ("format","format","format",
                      0, G_MAXINT, AFMT_S16_LE, G_PARAM_READWRITE)); 

  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_CHANNELS,
    g_param_spec_enum("channels","channels","channels",
                      GST_TYPE_COMEDISINK_CHANNELS,2,G_PARAM_READWRITE)); 
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_FREQUENCY,
    g_param_spec_int("frequency","frequency","frequency",
                     0,G_MAXINT,44100,G_PARAM_READWRITE));
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_FRAGMENT,
    g_param_spec_int("fragment","fragment","fragment",
                     0,G_MAXINT,6,G_PARAM_READWRITE));
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_BUFFER_SIZE,
    g_param_spec_int("buffer_size","buffer_size","buffer_size",
                     0,G_MAXINT,4096,G_PARAM_READWRITE));

  gst_comedisink_signals[SIGNAL_HANDOFF] =
    g_signal_new("handoff",G_TYPE_FROM_CLASS(klass), G_SIGNAL_RUN_LAST,
                   G_STRUCT_OFFSET(GstComediSinkClass,handoff), NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,G_TYPE_NONE,0);
  
  gobject_class->set_property = gst_comedisink_set_property;
  gobject_class->get_property = gst_comedisink_get_property;
  gobject_class->finalize     = gst_comedisink_finalize;
  
  gstelement_class->change_state = GST_DEBUG_FUNCPTR (gst_comedisink_change_state);
}

static GstClockTime 
gst_comedisink_get_time (GstClock *clock, gpointer data) 
{
  GstComediSink *comedisink = GST_COMEDISINK (data);
  gint delay;
  GstClockTime res;

  if (!comedisink->bps)
    return 0;

  ioctl (comedisink->fd, SNDCTL_DSP_GETODELAY, &delay);

  /* sometimes delay is bigger than the number of bytes sent to the device, which screws
   * up this calculation, we assume that everything is still in the device then */
  if (delay > comedisink->handled) {
    res = comedisink->offset;
  }
  else {
    res = comedisink->offset + (comedisink->handled - delay) * 1000000LL / comedisink->bps;
  }

  /*
  g_print ("from comedisink: %lld %lld %d %lld %d\n", res, comedisink->offset, delay, comedisink->handled,
		  comedisink->bps);
		  */

  return res;
}

static void 
gst_comedisink_init (GstComediSink *comedisink) 
{
  comedisink->sinkpad = gst_pad_new_from_template (
		  GST_PAD_TEMPLATE_GET (comedisink_sink_factory), "sink");
  gst_element_add_pad (GST_ELEMENT (comedisink), comedisink->sinkpad);
  gst_pad_set_connect_function (comedisink->sinkpad, gst_comedisink_sinkconnect);
  gst_pad_set_bufferpool_function (comedisink->sinkpad, gst_comedisink_get_bufferpool);

  gst_pad_set_chain_function (comedisink->sinkpad, gst_comedisink_chain);

  comedisink->device = g_strdup ("/dev/dsp");
  comedisink->fd = -1;
  comedisink->channels = 1;
  comedisink->frequency = 11025;
  comedisink->fragment = 6;
/* AFMT_*_BE not available on all COMEDI includes (e.g. FBSD) */
#ifdef WORDS_BIGENDIAN
  comedisink->format = AFMT_S16_BE;
#else
  comedisink->format = AFMT_S16_LE;
#endif /* WORDS_BIGENDIAN */  
  /* gst_clock_register (comedisink->clock, GST_OBJECT (comedisink)); */
  comedisink->bufsize = 4096;
  comedisink->bps = 0;
  comedisink->offset = 0LL;
  comedisink->handled = 0LL;
  /* 6 buffers per chunk by default */
  comedisink->sinkpool = gst_buffer_pool_get_default (comedisink->bufsize, 6);

  comedisink->provided_clock = NULL;
  comedisink->provided_clock = GST_CLOCK (gst_comedi_clock_new ("comediclock", gst_comedisink_get_time, comedisink));

  GST_FLAG_SET (comedisink, GST_ELEMENT_THREAD_SUGGESTED);
  GST_FLAG_SET (comedisink, GST_ELEMENT_EVENT_AWARE);
}

static GstPadConnectReturn 
gst_comedisink_sinkconnect (GstPad *pad, GstCaps *caps) 
{
  gint law, endianness, width, depth;
  gboolean sign;
  gint format = -1;
  GstComediSink *comedisink = GST_COMEDISINK (gst_pad_get_parent (pad));

  if (!GST_CAPS_IS_FIXED (caps))
    return GST_PAD_CONNECT_DELAYED;
  
  gst_caps_get_int (caps, "width", &width);
  gst_caps_get_int (caps, "depth", &depth);

  if (width != depth) 
    return GST_PAD_CONNECT_REFUSED;

  /* laws 1 and 2 are 1 bps anyway */
  comedisink->bps = 1;

  gst_caps_get_int (caps, "law", &law);
  gst_caps_get_int (caps, "endianness", &endianness);
  gst_caps_get_boolean (caps, "signed", &sign);

  if (law == 0) {
    if (width == 16) {
      if (sign == TRUE) {
        if (endianness == G_LITTLE_ENDIAN)
	  format = AFMT_S16_LE;
        else if (endianness == G_BIG_ENDIAN)
	  format = AFMT_S16_BE;
      }
      else {
        if (endianness == G_LITTLE_ENDIAN)
	  format = AFMT_U16_LE;
        else if (endianness == G_BIG_ENDIAN)
	  format = AFMT_U16_BE;
      }
      comedisink->bps = 2;
    }
    else if (width == 8) {
      if (sign == TRUE) {
	format = AFMT_S8;
      }
      else {
        format = AFMT_U8;
      }
      comedisink->bps = 1;
    }
  } else if (law == 1) {
    format = AFMT_MU_LAW;
  } else if (law == 2) {
    format = AFMT_A_LAW;
  } else {
    g_critical ("unknown law");
    return GST_PAD_CONNECT_REFUSED;
  }

  if (format == -1) 
    return GST_PAD_CONNECT_REFUSED;

  comedisink->format = format;
  gst_caps_get_int (caps, "channels", &comedisink->channels);
  gst_caps_get_int (caps, "rate", &comedisink->frequency);

  comedisink->bps *= comedisink->channels;
  comedisink->bps *= comedisink->frequency;

  if (!gst_comedisink_sync_parms (comedisink)) {
    return GST_PAD_CONNECT_REFUSED;
  }

  return GST_PAD_CONNECT_OK;
}

static gboolean 
gst_comedisink_sync_parms (GstComediSink *comedisink) 
{
  audio_buf_info ospace;
  int frag;
  gint target_format;
  gint target_channels;
  gint target_frequency;
  GObject *object;

  g_return_val_if_fail (comedisink != NULL, FALSE);
  g_return_val_if_fail (GST_IS_COMEDISINK (comedisink), FALSE);

  if (comedisink->fd == -1)
    return FALSE;
  
  if (comedisink->fragment >> 16)
    frag = comedisink->fragment;
  else
    frag = 0x7FFF0000 | comedisink->fragment;
  
  GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink: trying to set sound card to %dHz %d bit %s (%08x fragment)",
           comedisink->frequency, comedisink->format,
           (comedisink->channels == 2) ? "stereo" : "mono",frag);

  ioctl (comedisink->fd, SNDCTL_DSP_SETFRAGMENT, &frag);

  ioctl (comedisink->fd, SNDCTL_DSP_RESET, 0);

  target_format = comedisink->format;
  target_channels = comedisink->channels;
  target_frequency = comedisink->frequency;

  ioctl (comedisink->fd, SNDCTL_DSP_SETFMT, &comedisink->format);
  ioctl (comedisink->fd, SNDCTL_DSP_CHANNELS, &comedisink->channels);
  ioctl (comedisink->fd, SNDCTL_DSP_SPEED, &comedisink->frequency);

  ioctl (comedisink->fd, SNDCTL_DSP_GETBLKSIZE, &comedisink->fragment);
  ioctl (comedisink->fd, SNDCTL_DSP_GETOSPACE, &ospace);

  GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink: set sound card to %dHz %d bit %s (%d bytes buffer, %08x fragment)",
           comedisink->frequency, comedisink->format,
           (comedisink->channels == 2) ? "stereo" : "mono", ospace.bytes, comedisink->fragment);

  object = G_OBJECT (comedisink);
  g_object_freeze_notify (object);
  g_object_notify (object, "channels");
  g_object_notify (object, "frequency");
  g_object_notify (object, "fragment");
  g_object_notify (object, "format");
  g_object_thaw_notify (object);

  comedisink->fragment_time = (1000000 * comedisink->fragment) / comedisink->bps;
  GST_INFO (GST_CAT_PLUGIN_INFO, "fragment time %u %llu\n", comedisink->bps, comedisink->fragment_time);

  if (target_format != comedisink->format ||
      target_channels != comedisink->channels ||
      target_frequency != comedisink->frequency) 
  {
    g_warning ("could not configure comedi with required parameters, enjoy the noise :)");
    /* we could eventually return FALSE here, or just do some additional tests
     * to see that the frequencies don't differ too much etc.. */
  }
  return TRUE;
}

static void 
gst_comedisink_chain (GstPad *pad, GstBuffer *buf) 
{
  GstComediSink *comedisink;
  GstClockTime buftime;

  /* this has to be an audio buffer */
  comedisink = GST_COMEDISINK (gst_pad_get_parent (pad));

  if (GST_IS_EVENT (buf)) {
    switch (GST_EVENT_TYPE (buf)) {
      case GST_EVENT_EOS:
        ioctl (comedisink->fd, SNDCTL_DSP_SYNC);
	gst_pad_event_default (pad, GST_EVENT (buf));
        return;
      default:
        return;
    }
  }
  
  buftime = GST_BUFFER_TIMESTAMP (buf);

  if (!comedisink->bps) {
    gst_buffer_unref (buf);
    gst_element_error (GST_ELEMENT (comedisink), "capsnego was never performed, unknown data type");
  }

  if (comedisink->fd >= 0) {
    if (!comedisink->mute) {
      guchar *data = GST_BUFFER_DATA (buf);
      gint size = GST_BUFFER_SIZE (buf);
      /* gint frag = comedisink->fragment;   <-- unused, for reason see above */

      if (comedisink->clock) {
        /* FIXME, NEW_MEDIA/DISCONT?. Try to get our start point */
        if (!comedisink->have_offset && buftime != -1LL) {
	   GstClockTimeDiff jitter;

           GST_INFO (GST_CAT_PLUGIN_INFO, 
			   "comedisink: clock at offset: %lld, new offset %lld at time %lld\n", 
			   comedisink->offset, buftime, gst_clock_get_time (comedisink->clock));

	  comedisink->offset = buftime;
	  comedisink->have_offset = TRUE;
	  comedisink->handled = 0;
          gst_element_clock_wait (GST_ELEMENT (comedisink), comedisink->clock,
			 &jitter);
        }
	
	
	/* this doesn't work on BE machines, apparently
	while (size) {
          gint tosend = MIN (size, frag);
          write (comedisink->fd, data, tosend);
	  data += tosend;
	  size -= tosend;
          comedisink->handled += tosend;
	} */
	write (comedisink->fd, data, size);
        comedisink->handled += size;
      }
      /* no clock, try to be as fast as pcomediible */
      else {
        audio_buf_info ospace;

        ioctl (comedisink->fd, SNDCTL_DSP_GETOSPACE, &ospace);

        if (ospace.bytes >= size) {
          write (comedisink->fd, data, size);
	}
        comedisink->handled += size;
      }
    }
  }
  gst_buffer_unref (buf);
}

static void 
gst_comedisink_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec) 
{
  GstComediSink *comedisink;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_COMEDISINK (object));
  
  comedisink = GST_COMEDISINK (object);

  switch (prop_id) {
    case ARG_DEVICE:
      /* disallow changing the device while it is opened
         get_property("device") should return the right one */
      if (!GST_FLAG_IS_SET (comedisink, GST_COMEDISINK_OPEN))
      {
	g_free (comedisink->device);
        comedisink->device = g_strdup (g_value_get_string (value));
	g_object_notify (object, "device");
      }
      break;
    case ARG_MUTE:
      comedisink->mute = g_value_get_boolean (value);
      g_object_notify (G_OBJECT (comedisink), "mute");
      break;
    case ARG_FORMAT:
      comedisink->format = g_value_get_int (value);
      gst_comedisink_sync_parms (comedisink);
      break;
    case ARG_CHANNELS:
      comedisink->channels = g_value_get_enum (value);
      gst_comedisink_sync_parms (comedisink);
      break;
    case ARG_FREQUENCY:
      comedisink->frequency = g_value_get_int (value);
      gst_comedisink_sync_parms (comedisink);
      break;
    case ARG_FRAGMENT:
      comedisink->fragment = g_value_get_int (value);
      gst_comedisink_sync_parms (comedisink);
      break;
    case ARG_BUFFER_SIZE:
      if (comedisink->bufsize == g_value_get_int (value)) break;
      comedisink->bufsize = g_value_get_int (value);
      comedisink->sinkpool = gst_buffer_pool_get_default (comedisink->bufsize, 6);
      g_object_notify (object, "buffer_size");
      break;
    default:
      break;
  }
}

static void 
gst_comedisink_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec) 
{
  GstComediSink *comedisink;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_COMEDISINK (object));
  
  comedisink = GST_COMEDISINK (object);

  switch (prop_id) {
    case ARG_DEVICE:
      g_value_set_string (value, comedisink->device);
      break;
    case ARG_MUTE:
      g_value_set_boolean (value, comedisink->mute);
      break;
    case ARG_FORMAT:
      g_value_set_int (value, comedisink->format);
      break;
    case ARG_CHANNELS:
      g_value_set_enum (value, comedisink->channels);
      break;
    case ARG_FREQUENCY:
      g_value_set_int (value, comedisink->frequency);
      break;
    case ARG_FRAGMENT:
      g_value_set_int (value, comedisink->fragment);
      break;
    case ARG_BUFFER_SIZE:
      g_value_set_int (value, comedisink->bufsize);
      break;
    default:
      break;
  }
}

static gboolean
gst_comedisink_open_audio (GstComediSink *sink)
{
  gint caps;
  g_return_val_if_fail (sink->fd == -1, FALSE);

  GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink: attempting to open sound device");

  /* first try to open the sound card */
  sink->fd = open(sink->device, O_WRONLY | O_NONBLOCK);
  if (errno == EBUSY) {
    g_warning ("comedisink: unable to open the sound device (in use ?)\n");
    return FALSE;
  }

  /* re-open the sound device in blocking mode */
  close(sink->fd);
  sink->fd = open(sink->device, O_WRONLY);

  /* if we have it, set the default parameters and go have fun */
  if (sink->fd >= 0) {
    /* set card state */
    ioctl(sink->fd, SNDCTL_DSP_GETCAPS, &caps);

    GST_INFO(GST_CAT_PLUGIN_INFO, "comedisink: Capabilities %08x", caps);

    if (caps & DSP_CAP_DUPLEX)   	GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   Full duplex");
    if (caps & DSP_CAP_REALTIME) 	GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   Realtime");
    if (caps & DSP_CAP_BATCH)    	GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   Batch");
    if (caps & DSP_CAP_COPROC)   	GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   Has coprocessor");
    if (caps & DSP_CAP_TRIGGER)  	GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   Trigger");
    if (caps & DSP_CAP_MMAP)     	GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   Direct access");

#ifdef DSP_CAP_MULTI
    if (caps & DSP_CAP_MULTI)    	GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   Multiple open");
#endif /* DSP_CAP_MULTI */

#ifdef DSP_CAP_BIND
    if (caps & DSP_CAP_BIND)     	GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   Channel binding");
#endif /* DSP_CAP_BIND */

    ioctl(sink->fd, SNDCTL_DSP_GETFMTS, &caps);

    GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink: Formats %08x", caps);
    if (caps & AFMT_MU_LAW)   		GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   MU_LAW");
    if (caps & AFMT_A_LAW)    		GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   A_LAW");
    if (caps & AFMT_IMA_ADPCM)    	GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   IMA_ADPCM");
    if (caps & AFMT_U8)    		GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   U8");
    if (caps & AFMT_S16_LE)    		GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   S16_LE");
    if (caps & AFMT_S16_BE)    		GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   S16_BE");
    if (caps & AFMT_S8)    		GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   S8");
    if (caps & AFMT_U16_LE)    		GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   U16_LE");
    if (caps & AFMT_U16_BE)    		GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   U16_BE");
    if (caps & AFMT_MPEG)    		GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   MPEG");
#ifdef AFMT_AC3
    if (caps & AFMT_AC3)    		GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink:   AC3");
#endif

    GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink: opened audio (%s) with fd=%d", sink->device, sink->fd);
    GST_FLAG_SET (sink, GST_COMEDISINK_OPEN);

    return TRUE;
  }

  return FALSE;
}

static void
gst_comedisink_close_audio (GstComediSink *sink)
{
  if (sink->fd < 0) return;

  close(sink->fd);
  sink->fd = -1;

  GST_FLAG_UNSET (sink, GST_COMEDISINK_OPEN);

  GST_INFO (GST_CAT_PLUGIN_INFO, "comedisink: closed sound device");
}

static GstElementStateReturn 
gst_comedisink_change_state (GstElement *element) 
{
  GstComediSink *comedisink;

  g_return_val_if_fail (GST_IS_COMEDISINK (element), FALSE);

  comedisink = GST_COMEDISINK (element);

  switch (GST_STATE_TRANSITION (element)) {
    case GST_STATE_NULL_TO_READY:
      if (!GST_FLAG_IS_SET (element, GST_COMEDISINK_OPEN)) {
        if (!gst_comedisink_open_audio (comedisink)) {
          return GST_STATE_FAILURE;
        }
      }
      break;
    case GST_STATE_READY_TO_PAUSED:
      comedisink->offset = 0LL;
      comedisink->have_offset = FALSE;
      comedisink->handled = 0LL;
      break;
    case GST_STATE_PAUSED_TO_PLAYING:
      break;
    case GST_STATE_PLAYING_TO_PAUSED:
    {
      if (GST_FLAG_IS_SET (element, GST_COMEDISINK_OPEN)) 
        ioctl (comedisink->fd, SNDCTL_DSP_RESET, 0);
      break;
    }
    case GST_STATE_PAUSED_TO_READY:
      if (GST_FLAG_IS_SET (element, GST_COMEDISINK_OPEN))
        ioctl (comedisink->fd, SNDCTL_DSP_RESET, 0);
      break;
    case GST_STATE_READY_TO_NULL:
      if (GST_FLAG_IS_SET (element, GST_COMEDISINK_OPEN))
        gst_comedisink_close_audio (comedisink);
      break;
  }
      
  if (GST_ELEMENT_CLASS (parent_class)->change_state)
    return GST_ELEMENT_CLASS (parent_class)->change_state (element);

  return GST_STATE_SUCCESS;
}

gboolean 
gst_comedisink_factory_init (GstPlugin *plugin) 
{ 
  GstElementFactory *factory;

  factory = gst_element_factory_new ("comedisink", GST_TYPE_COMEDISINK, &gst_comedisink_details);
  g_return_val_if_fail (factory != NULL, FALSE);

  gst_element_factory_add_pad_template (factory, GST_PAD_TEMPLATE_GET (comedisink_sink_factory));

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  return TRUE;
}

