/* GStreamer
 * Copyright (C) 1999,2000 Erik Walthinsen <omega@cse.ogi.edu>
 *                    2000 Wim Taymans <wtay@chello.be>
 *                    2002 David A. Schleef <ds@schleef.org>
 *
 * gstcomedisrc.c: 
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
//#include <sys/soundcard.h>
//#include <sys/ioctl.h>
#include <unistd.h>
#include <comedilib.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>

#include <gstcomedisrc.h>

static GstElementDetails gst_comedisrc_details = {
	longname: "Comedi Source (COMEDI)",
	klass: "Source/Data",
	description: "Read from a Comedi device",
	version: VERSION,
	author: "David A. Schleef <ds@schleef.org>",
	copyright: "(C) 2002",
};


/* ComediSrc signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_DEVICE,
  ARG_BYTESPERREAD,
  ARG_CUROFFSET,
  ARG_FORMAT,
  ARG_CHANNELS,
  ARG_RATE,
};

GST_PAD_TEMPLATE_FACTORY (comedisrc_src_factory,
  "src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW (
    "comedisrc_src",
    "x-gst-sci/time-series",
      "format",		GST_PROPS_LIST (
	      		  GST_PROPS_STRING ("uint16"),
	      		  GST_PROPS_STRING ("float32"),
	      		  GST_PROPS_STRING ("float64")
			),
      "axis1",     	GST_PROPS_STRING("A/D voltage (V)"),
      "rate",     	GST_PROPS_INT_RANGE(1, 1000000000),
      "channels", 	GST_PROPS_INT_RANGE(1, 16)
  )
)

static void 			gst_comedisrc_class_init	(GstComediSrcClass *klass);
static void 			gst_comedisrc_init		(GstComediSrc *comedisrc);

static void 			gst_comedisrc_set_property	(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void 			gst_comedisrc_get_property	(GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static GstElementStateReturn 	gst_comedisrc_change_state	(GstElement *element);

static void 			gst_comedisrc_close		(GstComediSrc *src);
static gboolean 		gst_comedisrc_open		(GstComediSrc *src);
static void 			gst_comedisrc_sync_parms	(GstComediSrc *comedisrc);

static GstBuffer *		gst_comedisrc_get		(GstPad *pad);

static void dump_cmd(FILE *out,comedi_cmd *cmd);

static GstElementClass *parent_class = NULL;
/*static guint gst_comedisrc_signals[LAST_SIGNAL] = { 0 }; */

GType
gst_comedisrc_get_type (void) 
{
  static GType comedisrc_type = 0;

  if (!comedisrc_type) {
    static const GTypeInfo comedisrc_info = {
      sizeof(GstComediSrcClass),
      NULL,
      NULL,
      (GClassInitFunc)gst_comedisrc_class_init,
      NULL,
      NULL,
      sizeof(GstComediSrc),
      0,
      (GInstanceInitFunc)gst_comedisrc_init,
    };
    comedisrc_type = g_type_register_static (GST_TYPE_ELEMENT, "GstComediSrc", &comedisrc_info, 0);
  }
  return comedisrc_type;
}

static void
gst_comedisrc_class_init (GstComediSrcClass *klass) 
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_BYTESPERREAD,
    g_param_spec_ulong("bytes_per_read","bytes_per_read","bytes_per_read",
                       0,G_MAXULONG,0,G_PARAM_READWRITE)); /* CHECKME */
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_CUROFFSET,
    g_param_spec_ulong("curoffset","curoffset","curoffset",
                       0,G_MAXULONG,0,G_PARAM_READABLE)); /* CHECKME */
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_FORMAT,
    g_param_spec_int("format","format","format",
                     G_MININT,G_MAXINT,0,G_PARAM_READWRITE)); /* CHECKME */
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_CHANNELS,
    g_param_spec_int("channels","channels","channels",
                     G_MININT,G_MAXINT,0,G_PARAM_READWRITE)); /* CHECKME */
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_RATE,
    g_param_spec_int("rate","Rate","Acquisition rate",
                     0,G_MAXINT,10000,G_PARAM_READWRITE));
  g_object_class_install_property(G_OBJECT_CLASS(klass), ARG_DEVICE,
    g_param_spec_string("device","device","comedi device (/dev/comediN usually)",
                        "default",G_PARAM_READWRITE));
  
  gobject_class->set_property = gst_comedisrc_set_property;
  gobject_class->get_property = gst_comedisrc_get_property;

  gstelement_class->change_state = gst_comedisrc_change_state;
}

static void 
gst_comedisrc_init (GstComediSrc *comedisrc) 
{
  comedisrc->srcpad = gst_pad_new_from_template (
		  GST_PAD_TEMPLATE_GET (comedisrc_src_factory), "src");
  gst_pad_set_get_function(comedisrc->srcpad,gst_comedisrc_get);
  gst_element_add_pad (GST_ELEMENT (comedisrc), comedisrc->srcpad);

  comedisrc->device = g_strdup ("/dev/comedi0");
  comedisrc->fd = -1;

  /* adding some default values */
  comedisrc->format = 16; // XXX bogus
  comedisrc->channels = 1;
  comedisrc->rate = 10000;
  
  comedisrc->bytes_per_read = 4096;
  comedisrc->curoffset = 0;
  comedisrc->basetime = 0;
  comedisrc->samples_since_basetime = 0;
}

static inline void scaleconv_f32_u16(float *dest, uint16_t *src,
	unsigned int n, double offset, double increment)
{
	int i;

	for(i=0;i<n;i++){
		dest[i] = offset + increment * src[i];
	}
}

static GstBuffer *
gst_comedisrc_get (GstPad *pad)
{
  GstComediSrc *src;
  GstBuffer *buf;
  glong readsamples;
  glong readscans;
  int n_bytes;

  g_return_val_if_fail (pad != NULL, NULL);
  src = GST_COMEDISRC(gst_pad_get_parent (pad));

  GST_DEBUG (GST_CAT_PLUGIN_INFO, "attempting to read something from Comedi device");

  if (!GST_PAD_CAPS (pad)) {
    /* set caps on src pad */
    if (!gst_pad_try_set_caps (src->srcpad, 
		    GST_CAPS_NEW (
    		      "comedi_src",
		      "x-gst-sci/time-series",
        		"format",	GST_PROPS_STRING ("float32"),
		        "rate",		GST_PROPS_INT (src->rate),
      			"axis1",     	GST_PROPS_STRING("A/D voltage (V)"),
		        "channels",	GST_PROPS_INT (src->channels)
        	   ))) 
    {
      gst_element_error (GST_ELEMENT (src), "could not set caps");
      return NULL;
    }
  }

  if(src->read_buffer_len < src->bytes_per_read){
    src->read_buffer_len = src->bytes_per_read;
    src->read_buffer = realloc(src->read_buffer, src->read_buffer_len);
  }

  GST_DEBUG (GST_CAT_PLUGIN_INFO, "buffer %p length %d\n",src->read_buffer,src->read_buffer_len);
  GST_DEBUG (GST_CAT_PLUGIN_INFO, "bytes_per_read %ld\n",src->bytes_per_read);

  buf = gst_buffer_new ();
  g_return_val_if_fail (buf, NULL);
  
  n_bytes = 0;
  while(n_bytes < src->bytes_per_read){
	int readbytes;
	int save_errno;

  	comedi_poll(src->dev, src->subd);
  	readbytes = read (src->fd, src->read_buffer + n_bytes,
			src->bytes_per_read - n_bytes);
  	save_errno = errno;

  	GST_DEBUG (GST_CAT_PLUGIN_INFO, "read() returned %d\n",readbytes);

  	if (readbytes < 0) {
    		GST_ERROR (GST_ELEMENT(src), "read() returned errno=%d",save_errno);
		gst_element_set_eos (GST_ELEMENT (src));
    		return NULL;
  	}
#if 0
	if (readbytes == 0) {
		gst_element_set_eos (GST_ELEMENT (src));
		return NULL;
	}
#endif
	n_bytes += readbytes;
  }

  readscans = src->bytes_per_read / (sizeof(sampl_t) * src->channels);
  readsamples = readscans * src->channels;

  GST_BUFFER_DATA (buf) = (gpointer)g_malloc(readsamples * sizeof(float));
  GST_BUFFER_SIZE (buf) = readsamples * sizeof(float);
  GST_BUFFER_OFFSET (buf) = src->curoffset;
  GST_BUFFER_TIMESTAMP (buf) = src->basetime;

  scaleconv_f32_u16((float *)GST_BUFFER_DATA(buf), (uint16_t *)src->read_buffer,
	readsamples, -10.0, 20.0/65536);

  src->curoffset += src->bytes_per_read;
  src->basetime += readsamples * GST_NSECOND * src->interval;
  //src->samples_since_basetime += readsamples;

  GST_DEBUG (GST_CAT_PLUGIN_INFO, "pushed buffer from Comedi of %ld bytes, timestamp %" G_GUINT64_FORMAT, src->bytes_per_read, GST_BUFFER_TIMESTAMP (buf));
  return buf;
}

static void 
gst_comedisrc_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec) 
{
  GstComediSrc *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_COMEDISRC (object));
  
  src = GST_COMEDISRC (object);

  switch (prop_id) {
    case ARG_BYTESPERREAD:
      src->bytes_per_read = g_value_get_ulong (value);
      break;
    case ARG_FORMAT:
      src->format = g_value_get_int (value);
      break;
    case ARG_CHANNELS:
      src->channels = g_value_get_int (value);
      break;
    case ARG_RATE:
      /* Preserve the timestamps */
      src->basetime = src->samples_since_basetime * GST_NSECOND * src->interval;
      src->samples_since_basetime = 0;

      src->rate = g_value_get_int (value);
      break;
    case ARG_CUROFFSET:
      src->curoffset = g_value_get_ulong (value);
      break;
    case ARG_DEVICE:
      g_free(src->device);
      src->device = g_strdup (g_value_get_string (value));
      break;
    default:
      break;
  }
}

static void 
gst_comedisrc_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec) 
{
  GstComediSrc *src;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail (GST_IS_COMEDISRC (object));
  
  src = GST_COMEDISRC (object);

  switch (prop_id) {
    case ARG_BYTESPERREAD:
      g_value_set_ulong (value, src->bytes_per_read);
      break;
    case ARG_FORMAT:
      g_value_set_int (value, src->format);
      break;
    case ARG_CHANNELS:
      g_value_set_int (value, src->channels);
      break;
    case ARG_RATE:
      g_value_set_int (value, src->rate);
      break;
    case ARG_CUROFFSET:
      g_value_set_ulong (value, src->curoffset);
      break;
    case ARG_DEVICE:
      g_value_set_string (value, src->device);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static GstElementStateReturn 
gst_comedisrc_change_state (GstElement *element) 
{
  /* GstComediSrc *src = GST_COMEDISRC (element); */
  
  g_return_val_if_fail (GST_IS_COMEDISRC (element), FALSE);
  GST_DEBUG (GST_CAT_PLUGIN_INFO, "comedisrc: state change");
  /* if going down into NULL state, close the file if it's open */
  if (GST_STATE_PENDING (element) == GST_STATE_NULL) {
    if (GST_FLAG_IS_SET (element, GST_COMEDISRC_OPEN))
      gst_comedisrc_close (GST_COMEDISRC (element));
  /* otherwise (READY or higher) we need to open the sound card */
  } else {
    GST_DEBUG (GST_CAT_PLUGIN_INFO, "DEBUG: comedisrc: ready or higher");

    if (!GST_FLAG_IS_SET (element, GST_COMEDISRC_OPEN)) { 
      if (!gst_comedisrc_open (GST_COMEDISRC (element)))
        return GST_STATE_FAILURE;
      else
      {
	GST_DEBUG (GST_CAT_PLUGIN_INFO, "comedisrc: device opened successfully");
	/* thomas: we can't set caps here because the element is
	 * not actually ready yet */
      }
    }
  }

  if (GST_ELEMENT_CLASS (parent_class)->change_state)
    return GST_ELEMENT_CLASS (parent_class)->change_state (element);
  
  return GST_STATE_SUCCESS;
}

static gboolean 
gst_comedisrc_open (GstComediSrc *src) 
{
  g_return_val_if_fail (!GST_FLAG_IS_SET (src, GST_COMEDISRC_OPEN), FALSE);

  src->dev = comedi_open(src->device);

  if (src->dev == NULL) {
    return FALSE;
  }

  src->fd = comedi_fileno(src->dev);

  /* set card state */
  gst_comedisrc_sync_parms (src);
  GST_DEBUG (GST_CAT_PLUGIN_INFO,"opened: %s",src->device);
  
  GST_FLAG_SET (src, GST_COMEDISRC_OPEN);
  return TRUE;
}

static void 
gst_comedisrc_close (GstComediSrc *src) 
{
  g_return_if_fail (GST_FLAG_IS_SET (src, GST_COMEDISRC_OPEN));

  comedi_close(src->dev);
  src->dev = NULL;
  src->fd = -1;

  GST_FLAG_UNSET (src, GST_COMEDISRC_OPEN);
}

static void 
gst_comedisrc_sync_parms (GstComediSrc *comedisrc) 
{
  unsigned int interval;
  int ret;
  int i;

  g_return_if_fail (comedisrc != NULL);
  g_return_if_fail (GST_IS_COMEDISRC (comedisrc));
  g_return_if_fail (comedisrc->fd > 0);
 
  if(comedisrc->cmd_running)return;

  //subd = comedi_get_read_subdevice(comedisrc->dev);
  comedisrc->subd = 0;

  interval = (int)(1.0e9/comedisrc->rate);

  ret = comedi_get_cmd_generic_timed(comedisrc->dev, comedisrc->subd,
	&comedisrc->cmd, interval);
  if(ret<0){
    GST_DEBUG (GST_CAT_PLUGIN_INFO,"comedi_get_cmd_generic_timed() error %s",
	comedi_strerror(comedi_errno()));
  }

  for(i=0;i<16;i++){
    comedisrc->chanlist[i] = CR_PACK(i,0,AREF_DIFF);
  }

  comedisrc->cmd.chanlist = comedisrc->chanlist;
  comedisrc->cmd.chanlist_len = comedisrc->channels;

  comedisrc->cmd.scan_end_arg = comedisrc->channels;

  comedisrc->cmd.stop_src = TRIG_NONE;

  if(comedisrc->cmd.stop_src == TRIG_NONE)
    comedisrc->cmd.stop_arg = 0;
  if(comedisrc->cmd.stop_src == TRIG_COUNT)
    comedisrc->cmd.stop_arg = 1000000;

  ret = comedi_command_test(comedisrc->dev, &comedisrc->cmd);
  if(ret<0){
    GST_DEBUG (GST_CAT_PLUGIN_INFO,"comedi_command_test() error");
  }

  ret = comedi_command_test(comedisrc->dev, &comedisrc->cmd);
  if(ret<0){
    GST_DEBUG (GST_CAT_PLUGIN_INFO,"comedi_command_test() error");
  }
  if(ret!=0){
    GST_DEBUG (GST_CAT_PLUGIN_INFO,"comedi_command_test() problem");
  }

  dump_cmd(stdout,&comedisrc->cmd);
  comedisrc->interval = comedisrc->cmd.scan_begin_arg;

  ret = comedi_command(comedisrc->dev, &comedisrc->cmd);
  if(ret<0){
    GST_ERROR (GST_ELEMENT(comedisrc), "comedi_command_test() error");
  }

  comedisrc->cmd_running = 1;

  GST_DEBUG (GST_CAT_PLUGIN_INFO, "comedisrc: started command");
}

gboolean
gst_comedisrc_factory_init (GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_element_factory_new ("comedisrc", GST_TYPE_COMEDISRC, &gst_comedisrc_details);
  g_return_val_if_fail (factory != NULL, FALSE);

  gst_element_factory_add_pad_template (factory, GST_PAD_TEMPLATE_GET (comedisrc_src_factory));

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  return TRUE;
}

char *cmd_src(int src,char *buf)
{
	buf[0]=0;

	if(src&TRIG_NONE)strcat(buf,"none|");
	if(src&TRIG_NOW)strcat(buf,"now|");
	if(src&TRIG_FOLLOW)strcat(buf, "follow|");
	if(src&TRIG_TIME)strcat(buf, "time|");
	if(src&TRIG_TIMER)strcat(buf, "timer|");
	if(src&TRIG_COUNT)strcat(buf, "count|");
	if(src&TRIG_EXT)strcat(buf, "ext|");
	if(src&TRIG_INT)strcat(buf, "int|");
#ifdef TRIG_OTHER
	if(src&TRIG_OTHER)strcat(buf, "other|");
#endif

	if(strlen(buf)==0){
		sprintf(buf,"unknown(0x%08x)",src);
	}else{
		buf[strlen(buf)-1]=0;
	}

	return buf;
}

void dump_cmd(FILE *out,comedi_cmd *cmd)
{
	char buf[100];

	fprintf(out,"start:      %-8s %d\n",
		cmd_src(cmd->start_src,buf),
		cmd->start_arg);

	fprintf(out,"scan_begin: %-8s %d\n",
		cmd_src(cmd->scan_begin_src,buf),
		cmd->scan_begin_arg);

	fprintf(out,"convert:    %-8s %d\n",
		cmd_src(cmd->convert_src,buf),
		cmd->convert_arg);

	fprintf(out,"scan_end:   %-8s %d\n",
		cmd_src(cmd->scan_end_src,buf),
		cmd->scan_end_arg);

	fprintf(out,"stop:       %-8s %d\n",
		cmd_src(cmd->stop_src,buf),
		cmd->stop_arg);
}

