/* 
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 * Copyright (C) 2002      David A. Schleef <ds@schleef.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
#include <math.h>
#include <string.h>

/* include this header if you want to use dynamic parameters
#include <gst/control/control.h>
*/

#include "gstfftw.h"

static GstElementDetails plugin_details = {
	longname: "FFTWFilter",
	klass: "Filter/Fourier Transform",
	description: "Converts to/from frequency space",
	version: VERSION,
	author: "David A. Schleef <ds@schleef.org>",
	copyright: "(C) 2002"
};

/* Filter signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_SILENT,
  ARG_POWER,
  ARG_CHUNKSIZE,
};

GST_PAD_TEMPLATE_FACTORY (gst_plugin_template_sink_factory,
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "fft_sink",
    "x-gst-sci/time-series",
      "format",		GST_PROPS_STRING("float32"),
      "rate",		GST_PROPS_INT_RANGE(1, 1000000000),
      "channels",	GST_PROPS_INT(1)
  )
);

GST_PAD_TEMPLATE_FACTORY (gst_plugin_template_src_factory,
  "src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "fft_src",
    "x-gst-sci/frame",
      "axis0",		GST_PROPS_STRING("frequency (Hz)"),
      "axis0-offset",	GST_PROPS_FLOAT_RANGE(-INFINITY,INFINITY),
      "axis0-scale",	GST_PROPS_FLOAT_RANGE(-INFINITY,INFINITY),
      "axis1",		GST_PROPS_STRING("power (unknown)"),
      "format",		GST_PROPS_STRING("float32"),
      "rate",		GST_PROPS_INT_RANGE(1, 1000000000),
      "channels",	GST_PROPS_INT(1)
  )
);

static void	gst_plugin_template_class_init	(GstPluginTemplateClass *klass);
static void	gst_plugin_template_init		(GstPluginTemplate *filter);
static void gst_plugin_template_reinit(GstPluginTemplate *filter);

static void	gst_plugin_template_set_property	(GObject *object, guint prop_id,
                                                 const GValue *value, 
					         GParamSpec *pspec);
static void	gst_plugin_template_get_property    (GObject *object, guint prop_id,
                                                 GValue *value, 
						 GParamSpec *pspec);

static void	gst_plugin_template_chain		(GstPad *pad, GstBuffer *buf);

static GstElementClass *parent_class = NULL;

/* this function handles the connection with other plug-ins */
static GstPadConnectReturn
gst_plugin_sink_connect (GstPad *pad, GstCaps *caps)
{
  GstPluginTemplate *filter;
  double scale;
  int rate;
  
  filter = GST_PLUGIN_TEMPLATE (gst_pad_get_parent (pad));
  g_return_val_if_fail (filter != NULL, GST_PAD_CONNECT_REFUSED);
  g_return_val_if_fail (GST_IS_PLUGIN_TEMPLATE (filter), 
                        GST_PAD_CONNECT_REFUSED);
  
  if(!GST_CAPS_IS_FIXED (caps)) 
  {
    return GST_PAD_CONNECT_DELAYED;
  }

  gst_caps_get_int(caps,"rate",&rate);
  filter->rate = rate;
  filter->ns_per_sample = 1000000000/rate;
  scale = (double)rate/filter->chunk_size;

  return gst_pad_try_set_caps(filter->srcpad,
	GST_CAPS_NEW(
		"fftw_src",
    "x-gst-sci/frame",
      "axis0",		GST_PROPS_STRING("frequency (Hz)"),
      "axis0-offset",	GST_PROPS_FLOAT(0.0),
      "axis0-scale",	GST_PROPS_FLOAT(scale),
      "axis1",		GST_PROPS_STRING("power (unknown)"),
      "format",		GST_PROPS_STRING("float32"),
      "rate",		GST_PROPS_INT(1000),
      "channels",	GST_PROPS_INT(1)));
}

GType
gst_gst_fftw_filter_get_type (void) 
{
  static GType plugin_type = 0;

  if (!plugin_type) 
  {
    static const GTypeInfo plugin_info = 
    {
      sizeof (GstPluginTemplateClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_plugin_template_class_init,
      NULL,
      NULL,
      sizeof (GstPluginTemplate),
      0,
      (GInstanceInitFunc) gst_plugin_template_init,
    };
    plugin_type = g_type_register_static (GST_TYPE_ELEMENT, 
	                                  "GstFFTWFilter", 
	                                  &plugin_info, 0);
  }
  return plugin_type;
}

/* initialize the plugin's class */
static void
gst_plugin_template_class_init (GstPluginTemplateClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property (gobject_class, ARG_SILENT,
    g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
                          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_POWER,
    g_param_spec_boolean ("power", "Power", "Output power spectrum",
                          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_CHUNKSIZE,
    g_param_spec_int ("chunk_size", "Chunk size", "Chunk size",
    		1, 65536, 1024, G_PARAM_READWRITE));
  
  gobject_class->set_property = gst_plugin_template_set_property;
  gobject_class->get_property = gst_plugin_template_get_property;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_plugin_template_init (GstPluginTemplate *filter)
{
  filter->sinkpad = gst_pad_new_from_template (gst_plugin_template_sink_factory (), 
                                               "sink");
  gst_pad_set_connect_function (filter->sinkpad, gst_plugin_sink_connect);
  filter->srcpad = gst_pad_new_from_template (gst_plugin_template_src_factory (), 
                                              "src");
  //gst_pad_set_connect_function (filter->srcpad, gst_plugin_template_connect);
  
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_pad_set_chain_function (filter->sinkpad, gst_plugin_template_chain);
  filter->silent = FALSE;
  filter->power = FALSE;

  filter->chunk_size = 1024;
  filter->n_input = 0;

  gst_plugin_template_reinit(filter);
}

static void
gst_plugin_template_reinit(GstPluginTemplate *filter)
{
	/* FIXME free plan */

  filter->plan = rfftw_create_plan(filter->chunk_size, FFTW_FORWARD,
		FFTW_ESTIMATE);

  filter->data = g_realloc(filter->data,sizeof(double)*filter->chunk_size);
  filter->data_ft = g_realloc(filter->data_ft,sizeof(double)*filter->chunk_size);
}

/* chain function
 * this function does the actual processing
 */

static void conv_f64_f32(double *dest, float *src, unsigned int n)
{
	int i;
	for(i=0;i<n;i++){
		dest[i] = src[i];
	}
}

static void conv_f32_f64(float *dest, double *src, unsigned int n)
{
	int i;
	for(i=0;i<n;i++){
		dest[i] = src[i];
	}
}

#if 0
static void scalarmult_f32(float *dest, float *src, unsigned int n, double mult)
{
	int i;
	for(i=0;i<n;i++){
		dest[i] = src[i] * mult;
	}
}
#endif

static void powerize(float *dest, double *src, unsigned int n)
{
	int i;

	dest[0] = src[0]*src[0];
	for(i=1;i<n/2;i++){
		dest[i] = src[i]*src[i] + src[n-i]*src[n-i];
	}
}

static void window(double *dest, unsigned int n)
{
	double mult = 2*M_PI/n;
	int i;
	double x;

	for(i=0;i<n;i++){
		x = 0.5-0.5*cos(i*mult);
		dest[i] *= x;
	}
}

static void
gst_plugin_template_chain (GstPad *pad, GstBuffer *buf)
{
  GstPluginTemplate *filter;
  GstBuffer *outbuf;
  float *data;
  int n;
  int m;
  int orig_n_input;

  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);
  
  filter = GST_PLUGIN_TEMPLATE (GST_OBJECT_PARENT (pad));
  g_return_if_fail (GST_IS_PLUGIN_TEMPLATE (filter));

  if(GST_IS_EVENT(buf)){
	  g_print("Don't know how to handle events\n");
	  return;
  }

  data = (float *)GST_BUFFER_DATA(buf);
  n = GST_BUFFER_SIZE(buf)/sizeof(float);
  orig_n_input = filter->n_input;
  while(n>0){
	m = filter->chunk_size - filter->n_input;
	if(m>n) m = n;

	g_print("handling m=%d (n=%d, chunk=%d, n_input=%d)\n", m, n, filter->chunk_size, filter->n_input);

	conv_f64_f32(filter->data + filter->n_input, data, m);
	filter->n_input += m;
	data += m;
	n-=m;

	if(filter->n_input==filter->chunk_size){
		float *out_data;
		int out_len;

		window(filter->data, filter->chunk_size);
		rfftw_one(filter->plan, filter->data, filter->data_ft);

		outbuf = gst_buffer_new();
		if(filter->power){
			out_len = sizeof(float)*(filter->chunk_size/2);
		}else{
			out_len = sizeof(float)*filter->chunk_size;
		}
		out_data = g_malloc(out_len);
		GST_BUFFER_DATA(outbuf) = (void *)out_data;
		GST_BUFFER_SIZE(outbuf) = out_len;
		GST_BUFFER_TIMESTAMP(outbuf) = GST_BUFFER_TIMESTAMP(buf) +
			(filter->n_input - orig_n_input) * filter->ns_per_sample;

		if(filter->power){
			powerize(out_data, filter->data_ft, filter->chunk_size);
#if 0
			scalarmult_f32(out_data, out_data, filter->chunk_size,
					1.0/filter->chunk_size);
					//1.0/(filter->chunk_size*filter->chunk_size));
#endif
		}else{
			//conv_f32_f64(out_data, filter->data_ft, filter->chunk_size);
			conv_f32_f64(out_data, filter->data, filter->chunk_size);
		}

		gst_pad_push (filter->srcpad, outbuf);

		//filter->n_input = 0;

		memcpy(filter->data, filter->data+filter->chunk_size/2,
			filter->chunk_size/2*sizeof(double));
		filter->n_input = filter->chunk_size/2;
	}
  }
  gst_buffer_unref(buf);
}

static void
gst_plugin_template_set_property (GObject *object, guint prop_id, 
                              const GValue *value, GParamSpec *pspec)
{
  GstPluginTemplate *filter;

  g_return_if_fail (GST_IS_PLUGIN_TEMPLATE (object));
  filter = GST_PLUGIN_TEMPLATE (object);

  switch (prop_id) 
  {
  case ARG_SILENT:
    filter->silent = g_value_get_boolean (value);
    break;
  case ARG_POWER:
    filter->power = g_value_get_boolean (value);
    break;
  case ARG_CHUNKSIZE:
    filter->chunk_size = g_value_get_int (value);
    gst_plugin_template_reinit(filter);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gst_plugin_template_get_property (GObject *object, guint prop_id, 
                              GValue *value, GParamSpec *pspec)
{
  GstPluginTemplate *filter;

  g_return_if_fail (GST_IS_PLUGIN_TEMPLATE (object));
  filter = GST_PLUGIN_TEMPLATE (object);
  
  switch (prop_id) {
  case ARG_SILENT:
    g_value_set_boolean (value, filter->silent);
    break;
  case ARG_POWER:
    g_value_set_boolean (value, filter->power);
    break;
  case ARG_CHUNKSIZE:
    g_value_set_int (value, filter->chunk_size);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features 
 */
static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_element_factory_new ("fftw", GST_TYPE_PLUGIN_TEMPLATE,
                                     &plugin_details);
  g_return_val_if_fail (factory != NULL, FALSE);
  
  gst_element_factory_add_pad_template (factory, 
                                        gst_plugin_template_src_factory ());
  gst_element_factory_add_pad_template (factory, 
                                        gst_plugin_template_sink_factory ());

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  /* plugin initialisation succeeded */
  return TRUE;
}

/* this is the structure that gst-register looks for
 * so keep the name plugin_desc, or you cannot get your plug-in registered */
GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "fftw",
  plugin_init
};
