/* 
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 * Copyright (C) 2002      David A. Schleef <ds@schleef.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
#include <plplot/plplot.h>
#include <plplot/plplotP.h>
#include <math.h>

/* include this header if you want to use dynamic parameters
#include <gst/control/control.h>
*/

#include "gstplplotsink.h"

static GstElementDetails plugin_details = {
	longname: "PlplotSink",
	klass: "sink/plplot sink",
	description: "Plplot-based plotter",
	version: VERSION,
	author: "David A. Schleef <ds@schleef.org>",
	copyright: "(C) 2002"
};

/* Filter signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_SILENT,
  ARG_AUTOSCALE,
  ARG_TITLE,
  ARG_LOG,
  ARG_LOG_X,
  ARG_MARKER,
  ARG_Y_ZERO,
};

GST_PAD_TEMPLATE_FACTORY (gst_plugin_template_sink_factory,
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "plplotsink_sink",
    "x-gst-sci/frame",
      "format",		GST_PROPS_STRING("float32"),
      "rate",		GST_PROPS_INT_RANGE(1,1000000000),
      "channels",	GST_PROPS_INT_RANGE(1,16)
  )
);

static void gst_plugin_template_class_init	(GstPluginTemplateClass *klass);
static void gst_plugin_template_init		(GstPluginTemplate *filter);
static void gst_plplotsink_set_clock(GstElement *element, GstClock *clock);

static void gst_plugin_template_set_property	(GObject *object, guint prop_id,
                                                 const GValue *value, 
					         GParamSpec *pspec);
static void gst_plugin_template_get_property    (GObject *object, guint prop_id,
                                                 GValue *value, 
						 GParamSpec *pspec);

static void	gst_plugin_template_chain		(GstPad *pad, GstBuffer *buf);

static GstElementClass *parent_class = NULL;

/* this function handles the connection with other plug-ins */
static GstPadConnectReturn
gst_plugin_template_connect (GstPad *pad, GstCaps *caps)
{
  GstPluginTemplate *filter;
  
  filter = GST_PLUGIN_TEMPLATE (gst_pad_get_parent (pad));
  g_return_val_if_fail (filter != NULL, GST_PAD_CONNECT_REFUSED);
  g_return_val_if_fail (GST_IS_PLUGIN_TEMPLATE (filter), 
                        GST_PAD_CONNECT_REFUSED);
  
  if (GST_CAPS_IS_FIXED (caps)) 
  {
    /* caps on other side were accepted, so we're ok */
    return GST_PAD_CONNECT_OK;
  }
  /* not enough information yet, delay negotation */ 
  return GST_PAD_CONNECT_DELAYED;
}

GType
gst_gst_plugin_template_get_type (void) 
{
  static GType plugin_type = 0;

  if (!plugin_type) 
  {
    static const GTypeInfo plugin_info = 
    {
      sizeof (GstPluginTemplateClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_plugin_template_class_init,
      NULL,
      NULL,
      sizeof (GstPluginTemplate),
      0,
      (GInstanceInitFunc) gst_plugin_template_init,
    };
    plugin_type = g_type_register_static (GST_TYPE_ELEMENT, 
	                                  "GstPlplotSink", 
	                                  &plugin_info, 0);
  }
  return plugin_type;
}

/* initialize the plugin's class */
static void
gst_plugin_template_class_init (GstPluginTemplateClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property (gobject_class, ARG_SILENT,
    g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
                          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_AUTOSCALE,
    g_param_spec_boolean ("autoscale", "autoscale", "Automatically scale Y axis",
                          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_TITLE,
    g_param_spec_string ("title", "Title", "Plot Title",
	    "plplotsink", G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_LOG,
    g_param_spec_boolean ("log", "Logarithm", "Plot logarithm",
                          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_LOG_X,
    g_param_spec_boolean ("log_x", "Logarithm X axis", "Plot logarithm",
                          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_MARKER,
    g_param_spec_string ("marker", "Marker", "Marker",
	    "0.0,0.0", G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_Y_ZERO,
    g_param_spec_boolean ("zero", "Zero", "Always include 0 on y-axis",
                          FALSE, G_PARAM_READWRITE));
  
  gstelement_class->set_clock = gst_plplotsink_set_clock;

  gobject_class->set_property = gst_plugin_template_set_property;
  gobject_class->get_property = gst_plugin_template_get_property;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static int plugin_instance_count;

static void
gst_plugin_template_init (GstPluginTemplate *filter)
{
  filter->sinkpad = gst_pad_new_from_template (gst_plugin_template_sink_factory (), 
                                               "sink");
  gst_pad_set_connect_function (filter->sinkpad, gst_plugin_template_connect);
  
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_pad_set_chain_function (filter->sinkpad, gst_plugin_template_chain);
  filter->silent = FALSE;
  filter->autoscale = TRUE;
  filter->log = FALSE;
  filter->log_x = FALSE;
  filter->title = "plplotsink";

  filter->instance = plugin_instance_count++;

  plsstrm(filter->instance);
  plspage(0.0,0.0,640,480,0,0);
  plsdev("xwin");
  plinit();
  plspause(0);
  plssub(1,1);
  
  GST_FLAG_SET(filter, GST_ELEMENT_THREAD_SUGGESTED);
}

static void
gst_plplotsink_set_clock(GstElement *element, GstClock *clock)
{
	GstPluginTemplate *plplotsink;

	plplotsink = GST_PLUGIN_TEMPLATE(element);

	plplotsink->clock = clock;
}

/* chain function
 * this function does the actual processing
 */

static void
extents_f32(float *p_min, float *p_max, float *src, int n)
{
	float min, max;
	int i;

	min = max = src[0];
	for(i=1;i<n;i++){
		if(src[i]<min)min=src[i];
		if(src[i]>max)max=src[i];
	}
	*p_min = min;
	*p_max = max;
}

static void
gst_plugin_template_chain (GstPad *pad, GstBuffer *buf)
{
  GstPluginTemplate *filter;
  GstClockTime time, now;
  float *data;
  gint n;
  int i;
  GstCaps *caps;
  int rate;
  int channels;
  int chan;
  const gchar *x_axis;
  const gchar *y_axis;
  float x_axis_offset;
  float x_axis_scale;
  float y_min, y_max;

  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);
  
  filter = GST_PLUGIN_TEMPLATE (GST_OBJECT_PARENT (pad));
  g_return_if_fail (GST_IS_PLUGIN_TEMPLATE (filter));

  if(GST_IS_EVENT(buf)){
	  GstEvent *event = GST_EVENT(buf);
	  gint64 offset;

	  switch(GST_EVENT_TYPE(event)){
	    case GST_EVENT_DISCONTINUOUS:
		    offset = GST_EVENT_DISCONT_OFFSET(event,0).value;
		    g_print("plplot discont %" G_GINT64_FORMAT "\n", offset);
		    gst_clock_handle_discont(filter->clock,
			(guint64) GST_EVENT_DISCONT_OFFSET(event,0).value);
		    break;
	    default:
		gst_pad_event_default(pad,event);
		break;
	  }
	  gst_event_unref(event);
	  return;
  }

  time = GST_BUFFER_TIMESTAMP(buf);
  if(filter->clock && time != -1){
	  GstClockReturn ret;
  	  gint64 jitter;

	  GST_DEBUG(0,"plplotsink: clock %s wait: %" G_GUINT64_FORMAT " %u\n",
		GST_OBJECT_NAME(filter->clock), time,
		GST_BUFFER_SIZE(buf));
	  
	  now = gst_clock_get_time(filter->clock);
	  if(now >= time + filter->correction){
		  /* oops, we're already late.  let's increase
		   * correction */
		filter->correction = now - time;
		
  		if(!filter->silent){
			printf("increasing correction to %" G_GINT64_FORMAT "\n",now - time);
		}
	  }

	  if(filter->correction>1000000){
	  	filter->correction -= 1000000;
	  }

	  filter->id = gst_clock_new_single_shot_id(filter->clock, time);

	  ret = gst_clock_id_wait(filter->id, &jitter);
	  gst_clock_id_free(filter->id);
	  filter->id = NULL;
  }

  if(filter->clock){
	  time = gst_clock_get_time(filter->clock);
  }

  caps = gst_pad_get_caps(pad);
  gst_caps_get_int(caps, "rate", &rate);
  gst_caps_get_int(caps, "channels", &channels);
  gst_caps_get_string(caps, "axis0", &x_axis);
  gst_caps_get_string(caps, "axis1", &y_axis);
  gst_caps_get_float(caps, "axis0-offset", &x_axis_offset);
  gst_caps_get_float(caps, "axis0-scale", &x_axis_scale);

  data = (float *) GST_BUFFER_DATA(buf);
  n = GST_BUFFER_SIZE(buf)/sizeof(float)/channels;

  y_min = -10.0;
  y_max = 10.0;
  if(filter->autoscale){
	extents_f32(&y_min, &y_max, data, n);
//	y_min = pow(10.0, ceil(log10(fabs(y_min)))) * ((y_min<0)?-1:1);
//	y_max = pow(10.0, ceil(log10(fabs(y_max)))) * ((y_max<0)?-1:1);
	//if(y_min<filter->y_min)filter->y_min = y_min;
	//if(y_max>filter->y_max)filter->y_max = y_max;
	
	if(filter->y_max<y_max){
		filter->y_max = y_max;
	}else{
		filter->y_max = 0.99*filter->y_max + 0.01*y_max;
	}
	if(filter->y_min>y_min){
		filter->y_min = y_min;
	}else{
		filter->y_min = 0.99*filter->y_min + 0.01*y_min;
	}
	if(filter->y_zero){
		if(filter->y_min>0)filter->y_min = 0;
		if(filter->y_max<0)filter->y_max = 0;
	}
	y_min = filter->y_min;
	y_max = filter->y_max;
  }
  if(filter->log){
	  if(y_min>0){
		  y_min = log10(y_min);
	  }else{
		  y_min = -5;
	  }
	  if(y_max>0){
		  y_max = log10(y_max);
	  }else{
		  y_max = 0;
  	  }
	  if(y_min<y_max-10)y_min = y_max-10;
  }

  plsstrm(filter->instance);
  plbop();

  plcol(15);
  pladv(1);
  plvsta();
  if(filter->log_x){
  	plwind(log10(x_axis_offset + x_axis_scale), log10(x_axis_offset + x_axis_scale*n*0.5),
			y_min, y_max);
  }else{
  	plwind(x_axis_offset, x_axis_offset + x_axis_scale*n, y_min, y_max);
  }
  plbox("bcnst", 0.0, 0, "bcnstv", 0.0, 0);
  if(filter->log){
	char *y;

	y = g_strconcat("log ",y_axis);
  	pllab(x_axis, y, filter->title);
	free(y);
  }else{
  	pllab(x_axis, y_axis, filter->title);
  }

#define MAX_N 1000
  if(n<MAX_N){
	double *data_x, *data_y;

	data_x = malloc(sizeof(double)*n);
  	if(filter->log_x){
		for(i=0;i<n;i++){
			data_x[i] = log10(x_axis_offset + x_axis_scale * i);
		}
	}else{
		for(i=0;i<n;i++){
			data_x[i] = x_axis_offset + x_axis_scale * i;
		}
	}

	data_y = malloc(sizeof(double)*n);
	for(chan=0;chan<channels;chan++){
		if(filter->log){
			for(i=0;i<n;i++){
				double x;
				if(data[chan+channels*i]>0){
					x = log10(data[chan+channels*i]);
				}else{
					x = 0;
				}
				data_y[i] = x;
			}
		}else{
			for(i=0;i<n;i++){
				data_y[i] = data[chan+channels*i];
			}
		}

		plcol(1+chan);
		plline(n, data_x, data_y);
	}

	free(data_x);
	free(data_y);
  }else{
	double *data_x, *data_ymin, *data_ymax;
	double ratio = n/MAX_N;

	data_x = malloc(sizeof(double)*n);
  	if(filter->log_x){
		for(i=0;i<MAX_N;i++){
			data_x[i] = log10(x_axis_offset + x_axis_scale * i * ratio);
		}
	}else{
		for(i=0;i<MAX_N;i++){
			data_x[i] = x_axis_offset + x_axis_scale * i * ratio;
		}
	}

	data_ymin = malloc(sizeof(double)*MAX_N);
	data_ymax = malloc(sizeof(double)*MAX_N);
	for(chan=0;chan<channels;chan++){
		int last_j = -1;
		int j;
		double y;

		for(i=0;i<n;i++){
			j = floor(i*((double)MAX_N/n));
			y = data[chan+channels*i];
			if(j>last_j){
				data_ymin[j] = y;
				data_ymax[j] = y;
			}else{
				if(y<data_ymin[j])data_ymin[j] = y;
				if(y>data_ymax[j])data_ymax[j] = y;
			}
			last_j = j;
		}
		if(filter->log){
			for(i=0;i<MAX_N;i++){
				data_ymin[i] = log10(data_ymin[i]);
				data_ymax[i] = log10(data_ymax[i]);
			}
		}

		plcol(1+chan);
		plline(MAX_N, data_x, data_ymax);
		plline(MAX_N, data_x, data_ymin);
	}

	free(data_x);
	free(data_ymin);
	free(data_ymax);
  }

  if(filter->marker_show){
    plcol(2);
    plpoin(1, &filter->marker_x, &filter->marker_y, '+');
  }
  
  pleop();

#if 0
  if(filter->clock){
	  jitter = gst_clock_get_time(filter->clock) - time;

printf("clock jitter %" G_GINT64_FORMAT "\n",jitter);

	  //filter->correction = (filter->correction + jitter) >> 1;
	  filter->correction = (7*(filter->correction + jitter)) >> 4;
  }
#endif

  gst_buffer_unref(buf);
}

static void
gst_plugin_template_set_property (GObject *object, guint prop_id, 
                              const GValue *value, GParamSpec *pspec)
{
  GstPluginTemplate *filter;

  g_return_if_fail (GST_IS_PLUGIN_TEMPLATE (object));
  filter = GST_PLUGIN_TEMPLATE (object);

  switch (prop_id) 
  {
  case ARG_SILENT:
    filter->silent = g_value_get_boolean (value);
    break;
  case ARG_AUTOSCALE:
    filter->autoscale = g_value_get_boolean (value);
    break;
  case ARG_TITLE:
    filter->title = g_strdup(g_value_get_string (value));
    break;
  case ARG_LOG:
    filter->log = g_value_get_boolean (value);
    break;
  case ARG_LOG_X:
    filter->log_x = g_value_get_boolean (value);
    break;
  case ARG_MARKER: {
    const char *s;
    s = g_value_get_string (value);
    sscanf(s,"%lg,%lg",&filter->marker_x, &filter->marker_y);
    printf("marker at %g, %g\n", filter->marker_x, filter->marker_y);
    filter->marker_show = TRUE;
    }
    break;
  case ARG_Y_ZERO:
    filter->y_zero = g_value_get_boolean (value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gst_plugin_template_get_property (GObject *object, guint prop_id, 
                              GValue *value, GParamSpec *pspec)
{
  GstPluginTemplate *filter;

  g_return_if_fail (GST_IS_PLUGIN_TEMPLATE (object));
  filter = GST_PLUGIN_TEMPLATE (object);
  
  switch (prop_id) {
  case ARG_SILENT:
    g_value_set_boolean (value, filter->silent);
    break;
  case ARG_AUTOSCALE:
    g_value_set_boolean (value, filter->autoscale);
    break;
  case ARG_TITLE:
    g_value_set_string (value, filter->title);
    break;
  case ARG_LOG:
    g_value_set_boolean (value, filter->log);
    break;
  case ARG_LOG_X:
    g_value_set_boolean (value, filter->log_x);
    break;
  case ARG_Y_ZERO:
    g_value_set_boolean (value, filter->y_zero);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features 
 */
static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_element_factory_new ("plplotsink", GST_TYPE_PLUGIN_TEMPLATE,
                                     &plugin_details);
  g_return_val_if_fail (factory != NULL, FALSE);
  
  gst_element_factory_add_pad_template (factory, 
                                        gst_plugin_template_sink_factory ());

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  /* plugin initialisation succeeded */
  return TRUE;
}

/* this is the structure that gst-register looks for
 * so keep the name plugin_desc, or you cannot get your plug-in registered */
GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "plplotsink",
  plugin_init
};
