/* 
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 * Copyright (C) 2002      David A. Schleef <ds@schleef.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
//#include <gst/audio/audio.h>

/* include this header if you want to use dynamic parameters
#include <gst/control/control.h>
*/

#include <config.h>
#include "audio2data.h"

static GstElementDetails plugin_details = {
	longname: "Audio2Data",
	klass: "Generic/Audio2Data",
	description: "Convert audio to raw data",
	version: VERSION,
	author: "David A. Schleef <ds@schleef.org>",
	copyright: "(C) 2002"
};

/* Filter signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_SILENT
};

GST_PAD_TEMPLATE_FACTORY (gst_audio2data_sink_factory,
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "audio2data_sink",
    "audio/raw",
      "format",		GST_PROPS_STRING("int"),
      "law",		GST_PROPS_INT(0),
      "endianness",	GST_PROPS_INT(G_BYTE_ORDER),
      "signed",		GST_PROPS_BOOLEAN(TRUE),
      "width",		GST_PROPS_INT(16),
      "depth",		GST_PROPS_INT(16),
      "rate",		GST_PROPS_INT_RANGE(8000,96000),
      "channels",	GST_PROPS_INT(1)
));

GST_PAD_TEMPLATE_FACTORY (gst_audio2data_src_factory,
  "src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "audio2data_src",
    "x-gst-sci/time-series",
      "format",		GST_PROPS_STRING("float32"),
      "axis1",		GST_PROPS_STRING("sound input (level)"),
      "rate",		GST_PROPS_INT_RANGE(1,96000),
      "channels",	GST_PROPS_INT_RANGE(1,16)
  )
);

static void	gst_audio2data_class_init	(GstAudio2DataClass *klass);
static void	gst_audio2data_init		(GstAudio2Data *filter);

static void	gst_audio2data_set_property	(GObject *object, guint prop_id,
                                                 const GValue *value, 
					         GParamSpec *pspec);
static void	gst_audio2data_get_property    (GObject *object, guint prop_id,
                                                 GValue *value, 
						 GParamSpec *pspec);

static void	gst_audio2data_chain		(GstPad *pad, GstBuffer *buf);

static GstElementClass *parent_class = NULL;

/* this function handles the connection with other plug-ins */
static GstPadConnectReturn
gst_audio2data_connect_sink (GstPad *pad, GstCaps *caps)
{
  GstAudio2Data *filter;
  int rate;
  int channels;
  
  filter = GST_AUDIO2DATA (gst_pad_get_parent (pad));
  g_return_val_if_fail (filter != NULL, GST_PAD_CONNECT_REFUSED);
  g_return_val_if_fail (GST_IS_AUDIO2DATA (filter), 
                        GST_PAD_CONNECT_REFUSED);
  
  if (!GST_CAPS_IS_FIXED (caps)) 
  {
    return GST_PAD_CONNECT_DELAYED;
  }

  gst_caps_get_int(caps, "rate", &rate);
  gst_caps_get_int(caps, "channels", &channels);

  return gst_pad_try_set_caps(filter->srcpad,
  GST_CAPS_NEW(
    "audio2data_src",
    "x-gst-sci/time-series",
      "format",		GST_PROPS_STRING("float32"),
      "axis1",		GST_PROPS_STRING("sound input (level)"),
      "rate",		GST_PROPS_INT(rate),
      "channels",	GST_PROPS_INT(channels)
  ));
}

GType
gst_gst_audio2data_get_type (void) 
{
  static GType plugin_type = 0;

  if (!plugin_type) 
  {
    static const GTypeInfo plugin_info = 
    {
      sizeof (GstAudio2DataClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_audio2data_class_init,
      NULL,
      NULL,
      sizeof (GstAudio2Data),
      0,
      (GInstanceInitFunc) gst_audio2data_init,
    };
    plugin_type = g_type_register_static (GST_TYPE_ELEMENT, 
	                                  "GstAudio2Data", 
	                                  &plugin_info, 0);
  }
  return plugin_type;
}

/* initialize the plugin's class */
static void
gst_audio2data_class_init (GstAudio2DataClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property (gobject_class, ARG_SILENT,
    g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
                          FALSE, G_PARAM_READWRITE));
  
  gobject_class->set_property = gst_audio2data_set_property;
  gobject_class->get_property = gst_audio2data_get_property;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_audio2data_init (GstAudio2Data *filter)
{
  filter->sinkpad = gst_pad_new_from_template (gst_audio2data_sink_factory (), 
                                               "sink");
  gst_pad_set_connect_function (filter->sinkpad, gst_audio2data_connect_sink);
  filter->srcpad = gst_pad_new_from_template (gst_audio2data_src_factory (), 
                                              "src");
  
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_pad_set_chain_function (filter->sinkpad, gst_audio2data_chain);
  filter->silent = FALSE;

}

/* chain function
 * this function does the actual processing
 */

static void
conv_f32_s16(float *dest, short *src, int n)
{
	int i;
	for(i=0;i<n;i++){
		dest[i] = src[i];
	}
}

static void
gst_audio2data_chain (GstPad *pad, GstBuffer *buf)
{
  GstAudio2Data *filter;
  GstBuffer *outbuf;
  int samples;
  float *data;

  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);
  
  filter = GST_AUDIO2DATA (GST_OBJECT_PARENT (pad));
  g_return_if_fail (GST_IS_AUDIO2DATA (filter));

  samples = GST_BUFFER_SIZE(buf)/sizeof(short);

  outbuf = gst_buffer_new();
  g_return_if_fail(outbuf);

  data = g_malloc(samples*sizeof(float));
  GST_BUFFER_DATA(outbuf) = (gpointer)data;
  GST_BUFFER_SIZE(outbuf) = samples*sizeof(float);
  GST_BUFFER_TIMESTAMP(outbuf) = GST_BUFFER_TIMESTAMP(buf);

  conv_f32_s16(data, (short *)GST_BUFFER_DATA(buf), samples);

  gst_buffer_unref(buf);

  gst_pad_push(filter->srcpad, outbuf);
}

static void
gst_audio2data_set_property (GObject *object, guint prop_id, 
                              const GValue *value, GParamSpec *pspec)
{
  GstAudio2Data *filter;

  g_return_if_fail (GST_IS_AUDIO2DATA (object));
  filter = GST_AUDIO2DATA (object);

  switch (prop_id) 
  {
  case ARG_SILENT:
    filter->silent = g_value_get_boolean (value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gst_audio2data_get_property (GObject *object, guint prop_id, 
                              GValue *value, GParamSpec *pspec)
{
  GstAudio2Data *filter;

  g_return_if_fail (GST_IS_AUDIO2DATA (object));
  filter = GST_AUDIO2DATA (object);
  
  switch (prop_id) {
  case ARG_SILENT:
    g_value_set_boolean (value, filter->silent);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features 
 */
static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_element_factory_new ("audio2data", GST_TYPE_AUDIO2DATA,
                                     &plugin_details);
  g_return_val_if_fail (factory != NULL, FALSE);
  
  gst_element_factory_add_pad_template (factory, 
                                        gst_audio2data_src_factory ());
  gst_element_factory_add_pad_template (factory, 
                                        gst_audio2data_sink_factory ());

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  /* plugin initialisation succeeded */
  return TRUE;
}

/* this is the structure that gst-register looks for
 * so keep the name plugin_desc, or you cannot get your plug-in registered */
GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "audio",
  plugin_init
};
