/* 
 * 
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>

/* include this header if you want to use dynamic parameters
#include <gst/control/control.h>
*/

#include <config.h>
#include "decimate.h"

static GstElementDetails plugin_details = {
	longname: "Decimate",
	klass: "Generic/Decimate",
	description: "Decimates input",
	version: VERSION,
	author: "David A. Schleef <ds@schleef.org>",
	copyright: "(C) 2002"
};

/* Filter signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_SILENT,
  ARG_DECIMATION_FACTOR,
};

GST_PAD_TEMPLATE_FACTORY (gst_decimate_sink_factory,
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "decimate_sink",
    "x-gst-sci/time-series",
      "format",	GST_PROPS_STRING("float32"),
      "rate",	GST_PROPS_INT_RANGE(1, 1000000000),
      "channels",	GST_PROPS_INT(1)
  )
);

GST_PAD_TEMPLATE_FACTORY (gst_decimate_src_factory,
  "src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "decimate_sink",
    "x-gst-sci/time-series",
      "format",	GST_PROPS_STRING("float32"),
      "rate",	GST_PROPS_INT_RANGE(1, 1000000000),
      "channels",	GST_PROPS_INT(1)
  )
);

static void	gst_decimate_class_init	(GstDecimateClass *klass);
static void	gst_decimate_init		(GstDecimate *filter);

static void	gst_decimate_set_property	(GObject *object, guint prop_id,
                                                 const GValue *value, 
					         GParamSpec *pspec);
static void	gst_decimate_get_property    (GObject *object, guint prop_id,
                                                 GValue *value, 
						 GParamSpec *pspec);

static void	gst_decimate_chain		(GstPad *pad, GstBuffer *buf);

static GstElementClass *parent_class = NULL;

/* this function handles the connection with other plug-ins */
static GstPadConnectReturn
gst_decimate_connect (GstPad *pad, GstCaps *caps)
{
  GstDecimate *filter;
  int out_rate;
  
  filter = GST_DECIMATE (gst_pad_get_parent (pad));
  g_return_val_if_fail (filter != NULL, GST_PAD_CONNECT_REFUSED);
  g_return_val_if_fail (GST_IS_DECIMATE (filter), 
                        GST_PAD_CONNECT_REFUSED);
  
  if (!GST_CAPS_IS_FIXED (caps)) 
    return GST_PAD_CONNECT_DELAYED;

  gst_caps_get_int(caps,"rate",&out_rate);
  out_rate /= filter->decimation_factor;

  return gst_pad_try_set_caps (filter->srcpad, GST_CAPS_NEW(
    "decimate_src",
    "x-gst-sci/time-series",
      "format",	GST_PROPS_STRING("float32"),
      "rate",	GST_PROPS_INT(out_rate),
      "channels",	GST_PROPS_INT(1)
  ));
}

GType
gst_gst_decimate_get_type (void) 
{
  static GType plugin_type = 0;

  if (!plugin_type) 
  {
    static const GTypeInfo plugin_info = 
    {
      sizeof (GstDecimateClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_decimate_class_init,
      NULL,
      NULL,
      sizeof (GstDecimate),
      0,
      (GInstanceInitFunc) gst_decimate_init,
    };
    plugin_type = g_type_register_static (GST_TYPE_ELEMENT, 
	                                  "GstDecimate", 
	                                  &plugin_info, 0);
  }
  return plugin_type;
}

/* initialize the plugin's class */
static void
gst_decimate_class_init (GstDecimateClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property (gobject_class, ARG_SILENT,
    g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
                          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_DECIMATION_FACTOR,
    g_param_spec_int ("decimation_factor", "Decimation Factor", "Decimation factor",
	    0, G_MAXINT, 10, G_PARAM_READWRITE));
  
  gobject_class->set_property = gst_decimate_set_property;
  gobject_class->get_property = gst_decimate_get_property;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_decimate_init (GstDecimate *filter)
{
  filter->sinkpad = gst_pad_new_from_template (gst_decimate_sink_factory (), 
                                               "sink");
  gst_pad_set_connect_function (filter->sinkpad, gst_decimate_connect);
  filter->srcpad = gst_pad_new_from_template (gst_decimate_src_factory (), 
                                              "src");
  
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_pad_set_chain_function (filter->sinkpad, gst_decimate_chain);
  filter->silent = FALSE;

  filter->decimation_factor=10;
}

/* chain function
 * this function does the actual processing
 */

static void
gst_decimate_chain (GstPad *pad, GstBuffer *buf)
{
  GstDecimate *filter;
  GstBuffer *outbuf;
  float *out_data, *input_data;
  int n, i, index;
  int input_n;

  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);
  
  filter = GST_DECIMATE (GST_OBJECT_PARENT (pad));
  g_return_if_fail (GST_IS_DECIMATE (filter));

  outbuf = gst_buffer_new();

  input_n = GST_BUFFER_SIZE(buf)/sizeof(float);
  n = (input_n + filter->decimation_factor - 1 - filter->index) / filter->decimation_factor;
  input_data = (float *) GST_BUFFER_DATA(buf);

  //printf("index=%d in=%d n=%d\n", filter->index, input_n, n);

  out_data = g_malloc(n*sizeof(float));
  GST_BUFFER_SIZE(outbuf) = n*sizeof(float);
  GST_BUFFER_DATA(outbuf) = (gpointer) out_data;
  GST_BUFFER_TIMESTAMP(outbuf) = GST_BUFFER_TIMESTAMP(buf);

  index = filter->index;
  for(i=0;i<n;i++){
    out_data[i] = input_data[index];
    index += filter->decimation_factor;
  }
  filter->index = index - input_n;

  gst_pad_push (filter->srcpad, outbuf);
  gst_buffer_unref (buf);
}

static void
gst_decimate_set_property (GObject *object, guint prop_id, 
                              const GValue *value, GParamSpec *pspec)
{
  GstDecimate *filter;

  g_return_if_fail (GST_IS_DECIMATE (object));
  filter = GST_DECIMATE (object);

  switch (prop_id) 
  {
  case ARG_SILENT:
    filter->silent = g_value_get_boolean (value);
    break;
  case ARG_DECIMATION_FACTOR:
    filter->decimation_factor = g_value_get_int (value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gst_decimate_get_property (GObject *object, guint prop_id, 
                              GValue *value, GParamSpec *pspec)
{
  GstDecimate *filter;

  g_return_if_fail (GST_IS_DECIMATE (object));
  filter = GST_DECIMATE (object);
  
  switch (prop_id) {
  case ARG_SILENT:
    g_value_set_boolean (value, filter->silent);
    break;
  case ARG_DECIMATION_FACTOR:
    g_value_set_int (value, filter->decimation_factor);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features 
 */
static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_element_factory_new ("decimate", GST_TYPE_DECIMATE,
                                     &plugin_details);
  g_return_val_if_fail (factory != NULL, FALSE);
  
  gst_element_factory_add_pad_template (factory, 
                                        gst_decimate_src_factory ());
  gst_element_factory_add_pad_template (factory, 
                                        gst_decimate_sink_factory ());

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  /* plugin initialisation succeeded */
  return TRUE;
}

/* this is the structure that gst-register looks for
 * so keep the name plugin_desc, or you cannot get your plug-in registered */
GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "decimate",
  plugin_init
};
