/* 
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 * Copyright (C) 2002      David A. Schleef <ds@schleef.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
#include <math.h>

/* include this header if you want to use dynamic parameters
#include <gst/control/control.h>
*/

#include <config.h>
#include "envelope.h"

static GstElementDetails plugin_details = {
	longname:	"Envelope",
	klass:		"Filter/Envelope",
	description:	"Converts a waveform into its envelope",
	version:	VERSION,
	author:		"David A. Schleef <ds@schleef.org>",
	copyright:	"(C) 2002",
};

/* Filter signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_SILENT,
  ARG_TAP,
};

GST_PAD_TEMPLATE_FACTORY (gst_envelope_sink_factory,
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "envelope_sink",
    "x-gst-sci/time-series",
      "format",		GST_PROPS_STRING("float32"),
      "rate",		GST_PROPS_INT_RANGE(1, 1000000000),
      "channels",	GST_PROPS_INT(1)
  )
);

GST_PAD_TEMPLATE_FACTORY (gst_envelope_src_factory,
  "src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "envelope_src",
    "x-gst-sci/time-series",
      "format",		GST_PROPS_STRING("float32"),
      "rate",		GST_PROPS_INT_RANGE(1, 1000000000),
      "channels",	GST_PROPS_INT(1)
  )
);

static void	gst_envelope_class_init	(GstEnvelopeClass *klass);
static void	gst_envelope_init		(GstEnvelope *filter);

static void	gst_envelope_set_property	(GObject *object, guint prop_id,
                                                 const GValue *value, 
					         GParamSpec *pspec);
static void	gst_envelope_get_property    (GObject *object, guint prop_id,
                                                 GValue *value, 
						 GParamSpec *pspec);

static void	gst_envelope_chain		(GstPad *pad, GstBuffer *buf);

static GstElementClass *parent_class = NULL;

/* this function handles the connection with other plug-ins */
static GstPadConnectReturn
gst_envelope_connect (GstPad *pad, GstCaps *caps)
{
  GstEnvelope *filter;
  GstPad *otherpad;
  
  filter = GST_ENVELOPE (gst_pad_get_parent (pad));
  g_return_val_if_fail (filter != NULL, GST_PAD_CONNECT_REFUSED);
  g_return_val_if_fail (GST_IS_ENVELOPE (filter), 
                        GST_PAD_CONNECT_REFUSED);
  otherpad = (pad == filter->srcpad ? filter->sinkpad : filter->srcpad);
  
  if (GST_CAPS_IS_FIXED (caps)) 
  {
    /* caps are not fixed, so try to connect on the other side and see if
     * that works */

    if (!gst_pad_try_set_caps (otherpad, caps))
      return GST_PAD_CONNECT_REFUSED;

    /* caps on other side were accepted, so we're ok */
    return GST_PAD_CONNECT_OK;
  }
  /* not enough information yet, delay negotation */ 
  return GST_PAD_CONNECT_DELAYED;
}

GType
gst_gst_envelope_get_type (void) 
{
  static GType plugin_type = 0;

  if (!plugin_type) 
  {
    static const GTypeInfo plugin_info = 
    {
      sizeof (GstEnvelopeClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_envelope_class_init,
      NULL,
      NULL,
      sizeof (GstEnvelope),
      0,
      (GInstanceInitFunc) gst_envelope_init,
    };
    plugin_type = g_type_register_static (GST_TYPE_ELEMENT, 
	                                  "GstEnvelope", 
	                                  &plugin_info, 0);
  }
  return plugin_type;
}

/* initialize the plugin's class */
static void
gst_envelope_class_init (GstEnvelopeClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property (gobject_class, ARG_SILENT,
    g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
                          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_TAP,
    g_param_spec_float ("tap", "Tap", "Tap",
	    0.0, 1.0, 0.01, G_PARAM_READWRITE));
  
  gobject_class->set_property = gst_envelope_set_property;
  gobject_class->get_property = gst_envelope_get_property;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_envelope_init (GstEnvelope *filter)
{
  filter->sinkpad = gst_pad_new_from_template (gst_envelope_sink_factory (), 
                                               "sink");
  gst_pad_set_connect_function (filter->sinkpad, gst_envelope_connect);
  filter->srcpad = gst_pad_new_from_template (gst_envelope_src_factory (), 
                                              "src");
  //gst_pad_set_connect_function (filter->srcpad, gst_envelope_connect);
  
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_pad_set_chain_function (filter->sinkpad, gst_envelope_chain);
  filter->silent = FALSE;

  filter->srctap = 0.01;
  filter->tap0 = 0.99;

}

/* chain function
 * this function does the actual processing
 */

static void square_f32(float *dest, float *src, int n)
{
	int i;

	for(i=0;i<n;i++){
		dest[i] = src[i] * src[i];
	}
}

static void iir1_f32(float *dest, float *state, float *src, int n, float srctap, float tap0)
{
	int i;
	double x;

	x = state[0];
	for(i=0;i<n;i++){
		x = x * tap0 + src[i] * srctap;
		dest[i] = sqrt(x);
	}
	state[0] = x;
}

static void gst_envelope_filter(GstEnvelope *envelope, float *data, int n)
{
	square_f32(data, data, n);

	iir1_f32(data, envelope->state, data, n, envelope->srctap,
		envelope->tap0);
}

static void
gst_envelope_chain (GstPad *pad, GstBuffer *buf)
{
  GstEnvelope *filter;

  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);
  
  filter = GST_ENVELOPE (GST_OBJECT_PARENT (pad));
  g_return_if_fail (GST_IS_ENVELOPE (filter));

  buf = gst_buffer_copy_on_write(buf);

  gst_envelope_filter(filter, (float *)GST_BUFFER_DATA(buf),
	GST_BUFFER_SIZE(buf)/sizeof(float));

  gst_pad_push (filter->srcpad, buf);
}

static void
gst_envelope_set_property (GObject *object, guint prop_id, 
                              const GValue *value, GParamSpec *pspec)
{
  GstEnvelope *filter;

  g_return_if_fail (GST_IS_ENVELOPE (object));
  filter = GST_ENVELOPE (object);

  switch (prop_id) 
  {
  case ARG_SILENT:
    filter->silent = g_value_get_boolean (value);
    break;
  case ARG_TAP:
    filter->srctap = g_value_get_float (value);
    filter->tap0 = 1.0 - filter->srctap;
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gst_envelope_get_property (GObject *object, guint prop_id, 
                              GValue *value, GParamSpec *pspec)
{
  GstEnvelope *filter;

  g_return_if_fail (GST_IS_ENVELOPE (object));
  filter = GST_ENVELOPE (object);
  
  switch (prop_id) {
  case ARG_SILENT:
    g_value_set_boolean (value, filter->silent);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features 
 */
static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_element_factory_new ("envelope", GST_TYPE_ENVELOPE, 
                                     &plugin_details);
  g_return_val_if_fail (factory != NULL, FALSE);
  
  gst_element_factory_add_pad_template (factory, 
                                        gst_envelope_src_factory ());
  gst_element_factory_add_pad_template (factory, 
                                        gst_envelope_sink_factory ());

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  /* plugin initialisation succeeded */
  return TRUE;
}

/* this is the structure that gst-register looks for
 * so keep the name plugin_desc, or you cannot get your plug-in registered */
GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "envelope", 
  plugin_init
};
