/* 
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 * Copyright (C) 2002      David A. Schleef <ds@schleef.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
#include <math.h>
#include <string.h>

#define GST_SECOND_DOUBLE ((double)GST_SECOND)
#define GST_SECOND_MULDIV(a,b) (((a)*GST_SECOND_DOUBLE)/(b))
#define GST_SECOND_DIV(b) (GST_SECOND_DOUBLE/(b))

/* include this header if you want to use dynamic parameters
#include <gst/control/control.h>
*/

//#define DEBUG printf
#define DEBUG(a...)

#include <config.h>
#include "framer.h"

static GstElementDetails plugin_details = {
  longname: "Framer",
  klass: "Generic/Framer",
  description: "Create frames of input data",
  version: VERSION,
  author: "David Schleef <ds@schleef.org>",
  copyright: "(C) 2002"
};

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0,
  ARG_SILENT,
  ARG_CHUNK_SIZE,
  ARG_MAX_RATE,
  ARG_TRIG_OFFSET,
};

GST_PAD_TEMPLATE_FACTORY (gst_framer_sink_factory,
			  "sink",
			  GST_PAD_SINK,
			  GST_PAD_ALWAYS,
			  GST_CAPS_NEW ("framer_sink",
					"x-gst-sci/time-series",
					"format", GST_PROPS_STRING ("float32"),
					"rate", GST_PROPS_INT_RANGE (1,
								     1000000000),
					"channels", GST_PROPS_INT (1)
			  )
  );

GST_PAD_TEMPLATE_FACTORY (gst_framer_sink2_factory,
			  "sink2",
			  GST_PAD_SINK,
			  GST_PAD_ALWAYS,
			  GST_CAPS_NEW ("framer_trigger_sink",
					"x-gst-sci/timestamp-series",
                                        NULL
			  )
  );

GST_PAD_TEMPLATE_FACTORY (gst_framer_src_factory,
			  "src",
			  GST_PAD_SRC,
			  GST_PAD_ALWAYS,
			  GST_CAPS_NEW ("framer_src",
					"x-gst-sci/frame",
					"format", GST_PROPS_STRING ("float32"),
					"axis0", GST_PROPS_STRING ("time (s)"),
					"axis0-offset",
					GST_PROPS_FLOAT_RANGE (-INFINITY,
							       INFINITY),
					"axis0-scale",
					GST_PROPS_FLOAT_RANGE (-INFINITY,
							       INFINITY),
					"rate", GST_PROPS_INT_RANGE (1,
								     1000000000),
					"channels", GST_PROPS_INT (1)
			  )
  );

static void gst_framer_class_init (GstFramerClass * klass);
static void gst_framer_init (GstFramer * filter);

static void gst_framer_set_property (GObject * object, guint prop_id,
				     const GValue * value, GParamSpec * pspec);
static void gst_framer_get_property (GObject * object, guint prop_id,
				     GValue * value, GParamSpec * pspec);

static void gst_framer_loop_triggered (GstFramer *filter);
static void gst_framer_loop_nontriggered (GstFramer *filter);
static void gst_framer_loop (GstElement * element);
static void gst_framer_chain (GstPad * pad, GstBuffer * buf);

static GstElementClass *parent_class = NULL;

/* this function handles the connection with other plug-ins */
static GstPadConnectReturn
gst_framer_sink_connect (GstPad * pad, GstCaps * caps)
{
  GstFramer *filter;
  const gchar *axis1;
  int out_rate;
  int rate;

  filter = GST_FRAMER (gst_pad_get_parent (pad));
  g_return_val_if_fail (filter != NULL, GST_PAD_CONNECT_REFUSED);
  g_return_val_if_fail (GST_IS_FRAMER (filter), GST_PAD_CONNECT_REFUSED);

  if (!GST_CAPS_IS_FIXED (caps)) {
    return GST_PAD_CONNECT_DELAYED;
  }

  g_return_val_if_fail (filter->srcpad != NULL, GST_PAD_CONNECT_DELAYED);

  GST_DEBUG (GST_CAT_PLUGIN_INFO, "got here 1");

  gst_caps_get_int (caps, "rate", &rate);
  filter->rate = rate;
printf("rate=%g\n",filter->rate);

  if(gst_caps_has_property(caps, "axis1")){
    gst_caps_get_string (caps, "axis1", &axis1);
  }else{
    axis1 = "somebody is missing axis1";
  }

  out_rate = filter->rate / filter->chunk_size;
  if (out_rate < 1)
    out_rate = 1;

  GST_DEBUG (GST_CAT_PLUGIN_INFO, "got here 2");

  return gst_pad_try_set_caps (filter->srcpad,
			       GST_CAPS_NEW ("framer_src",
					     "x-gst-sci/frame",
					     "format",
					     GST_PROPS_STRING ("float32"),
					     "axis0",
					     GST_PROPS_STRING ("time (s)"),
					     "axis0-offset",
					     GST_PROPS_FLOAT (0.0),
					     "axis0-scale",
					     GST_PROPS_FLOAT ((1.0 /
							       filter->rate)),
					     "axis1",
					     GST_PROPS_STRING (g_strdup
							       (axis1)), "rate",
					     GST_PROPS_INT (out_rate),
					     "channels", GST_PROPS_INT (1)
			       ));
}

static GstPadConnectReturn
gst_framer_sink2_connect (GstPad * pad, GstCaps * caps)
{
  GstFramer *filter;

  filter = GST_FRAMER (gst_pad_get_parent (pad));
  g_return_val_if_fail (filter != NULL, GST_PAD_CONNECT_REFUSED);
  g_return_val_if_fail (GST_IS_FRAMER (filter), GST_PAD_CONNECT_REFUSED);

  if (!GST_CAPS_IS_FIXED (caps)) {
    return GST_PAD_CONNECT_DELAYED;
  }

  filter->triggered = TRUE;

  return GST_PAD_CONNECT_OK;
}

GType
gst_gst_framer_get_type (void)
{
  static GType plugin_type = 0;

  if (!plugin_type) {
    static const GTypeInfo plugin_info = {
      sizeof (GstFramerClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_framer_class_init,
      NULL,
      NULL,
      sizeof (GstFramer),
      0,
      (GInstanceInitFunc) gst_framer_init,
    };

    plugin_type =
      g_type_register_static (GST_TYPE_ELEMENT, "GstFramer", &plugin_info, 0);
  }
  return plugin_type;
}

/* initialize the plugin's class */
static void
gst_framer_class_init (GstFramerClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property (gobject_class, ARG_SILENT,
				   g_param_spec_boolean ("silent", "Silent",
							 "Produce verbose output ?",
							 FALSE,
							 G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_CHUNK_SIZE,
				   g_param_spec_int ("chunk_size", "Chunk size",
						     "Chunk Size", 1, G_MAXINT,
						     1024, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_MAX_RATE,
    g_param_spec_int ("max_rate", "Max rate", "Maximum output rate",
	    0, G_MAXINT, 100, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_TRIG_OFFSET,
    g_param_spec_int ("trigger_offset", "Trigger offset", "Trigger offset",
	    G_MININT, G_MAXINT, 0, G_PARAM_READWRITE));

  gobject_class->set_property = gst_framer_set_property;
  gobject_class->get_property = gst_framer_get_property;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_framer_init (GstFramer * filter)
{
  filter->sinkpad =
    gst_pad_new_from_template (gst_framer_sink_factory (), "sink");
  gst_pad_set_connect_function (filter->sinkpad, gst_framer_sink_connect);
  filter->trigger_sinkpad =
    gst_pad_new_from_template (gst_framer_sink2_factory (), "trigger_sink");
  gst_pad_set_connect_function (filter->trigger_sinkpad,
				gst_framer_sink2_connect);
  filter->srcpad = gst_pad_new_from_template (gst_framer_src_factory (), "src");

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->trigger_sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  //gst_pad_set_chain_function (filter->sinkpad, gst_framer_chain);
  (void) gst_framer_chain;
  gst_element_set_loop_function (GST_ELEMENT (filter), gst_framer_loop);
  filter->silent = FALSE;
  filter->chunk_size = 1024;

  filter->triggered = FALSE;
}

/* chain function
 * this function does the actual processing
 */

static void
gst_framer_chain (GstPad * pad, GstBuffer * buf)
{
  GstFramer *filter;
  float *data;
  int orig_n_input;
  int n;
  int len;

  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);

  filter = GST_FRAMER (GST_OBJECT_PARENT (pad));
  g_return_if_fail (GST_IS_FRAMER (filter));

  data = (void *) GST_BUFFER_DATA (buf);
  orig_n_input = filter->n_input;
  len = GST_BUFFER_SIZE (buf) / sizeof (float);
  n = len;
  while (n > 0) {
    int m;

    m = filter->chunk_size - filter->n_input;
    if (m > n)
      m = n;

    g_print ("handling m=%d (n=%d, chunk=%d, n_input=%d)\n",
	     m, n, filter->chunk_size, filter->n_input);

    if (filter->outbuf == NULL) {
      filter->outbuf = gst_buffer_new ();
      filter->data = g_malloc (sizeof (float) * filter->chunk_size);
      GST_BUFFER_DATA (filter->outbuf) = (void *) filter->data;
      GST_BUFFER_SIZE (filter->outbuf) = sizeof (float) * filter->chunk_size;
      GST_BUFFER_TIMESTAMP (filter->outbuf) = GST_BUFFER_TIMESTAMP (buf) +
	GST_SECOND_MULDIV(len - n, filter->rate);
    }

    memcpy (filter->data + filter->n_input, data, m * sizeof (float));
    filter->n_input += m;
    data += m;
    n -= m;

    if (filter->n_input == filter->chunk_size) {
      gst_pad_push (filter->srcpad, filter->outbuf);

      filter->data = NULL;
      filter->outbuf = NULL;
      filter->n_input = 0;
    }

  }

  gst_buffer_unref (buf);
}

static void
gst_framer_loop (GstElement * element)
{
  GstFramer *filter;

  filter = GST_FRAMER (element);
  g_return_if_fail (GST_IS_FRAMER (filter));

  if(GST_PAD_IS_CONNECTED(filter->trigger_sinkpad)){
    gst_framer_loop_triggered(filter);
  }else{
    gst_framer_loop_nontriggered(filter);
  }
}

static void
gst_framer_loop_nontriggered (GstFramer *filter)
{
  GstBuffer *input_buf;
  GstBuffer *output_buf;
  float *output_data;
  float *input_data;
  int n;
  int min_index;

  g_return_if_fail (GST_IS_FRAMER (filter));

  if(!filter->out_data){
    filter->out_data = malloc(filter->chunk_size * sizeof(float));
    memset(filter->out_data,0,filter->chunk_size * sizeof(float));
  }

  input_buf = gst_pad_pull (filter->sinkpad);
  n = GST_BUFFER_SIZE(input_buf)/sizeof(float);

  input_data = (float *) GST_BUFFER_DATA(input_buf);

  if(n<filter->chunk_size){
    memmove(filter->out_data, filter->out_data + n,
      (filter->chunk_size - n) * sizeof(float));
    memcpy(filter->out_data + (filter->chunk_size - n),
      input_data, n * sizeof(float));
  }else{
    memcpy(filter->out_data,
      input_data + (n - filter->chunk_size), filter->chunk_size * sizeof(float));
  }

  filter->out_index += n;

  if(filter->max_rate){
    min_index = filter->rate/filter->max_rate;
  }else{
    min_index = filter->chunk_size/20;
  }
  if(filter->out_index >= min_index){
    filter->out_index = 0;
    output_buf = gst_buffer_new();
    output_data = g_malloc(filter->chunk_size * sizeof (float));
    memcpy(output_data, filter->out_data, filter->chunk_size * sizeof(float));
    GST_BUFFER_DATA(output_buf) = (gpointer) output_data;
    GST_BUFFER_SIZE(output_buf) = filter->chunk_size * sizeof(float);

    GST_BUFFER_TIMESTAMP(output_buf) = GST_BUFFER_TIMESTAMP(input_buf);

    gst_pad_push (filter->srcpad, output_buf);
  }

  gst_buffer_unref (input_buf);

}

static void
gst_framer_loop_triggered (GstFramer *filter)
{
  int offset = 0;
  int do_copy = 0;

  g_return_if_fail (GST_IS_FRAMER (filter));

  DEBUG("filter time=%lld trig_time=%lld\n", filter->time, filter->trig_time);

  if (!filter->trig_buf && filter->triggered) {
    guint64 newtime;

    filter->trig_buf = gst_pad_pull (filter->trigger_sinkpad);
    DEBUG("pulling trigger %lld\n", GST_BUFFER_TIMESTAMP(filter->trig_buf));

    filter->trig_data = (guint64 *)GST_BUFFER_DATA(filter->trig_buf);

    filter->trig_index = 0;
    filter->trig_n = GST_BUFFER_SIZE (filter->trig_buf) / sizeof (guint64);

    if(filter->trig_n>0){
      newtime = filter->trig_data[0] + filter->trig_offset;
    }else{
      newtime = GST_BUFFER_TIMESTAMP (filter->trig_buf) + filter->trig_offset;
    }
    if(newtime > filter->trig_time)filter->trig_time = newtime;
  }

  if (!filter->input_buf){
    filter->input_buf = gst_pad_pull (filter->sinkpad);

    filter->input_n = GST_BUFFER_SIZE (filter->input_buf) / sizeof (float);
    filter->input_endtime = GST_BUFFER_TIMESTAMP (filter->input_buf) +
      GST_SECOND_MULDIV(filter->input_n, filter->rate);
    filter->input_data = (float *)GST_BUFFER_DATA(filter->input_buf);
    DEBUG("pulling input %lld -> %lld\n", GST_BUFFER_TIMESTAMP(filter->input_buf), filter->input_endtime);
  }

  if (!filter->out_buf) {
    DEBUG("new out_buf\n");
    filter->out_buf = gst_buffer_new ();
    filter->out_data = g_malloc (filter->chunk_size * sizeof (float));

    GST_BUFFER_SIZE (filter->out_buf) = filter->chunk_size * sizeof (float);
    GST_BUFFER_DATA (filter->out_buf) = (void *) filter->out_data;

    filter->out_index = 0;
  }

  //if(!filter->in_trigger){
  {
    DEBUG("skipping1\n");
    while (filter->trig_index < filter->trig_n &&
        filter->trig_data[filter->trig_index] + filter->trig_offset < filter->trig_time){
      filter->trig_index++;
    }
  }

  if(!filter->in_trigger){
    if (filter->trig_index < filter->trig_n || !filter->triggered){
      guint64 trigtime;
      if(filter->triggered){
	trigtime = filter->trig_data[filter->trig_index] + filter->trig_offset;
      }else{
	trigtime = filter->trig_time;
      }
      if(trigtime < filter->input_endtime){
if(filter->triggered){
DEBUG("triggering on index %d (%d)\n",filter->trig_index, filter->trig_n);
DEBUG("trigger input buf %lld -> %lld\n", GST_BUFFER_TIMESTAMP(filter->input_buf), filter->input_endtime);
DEBUG("trigger trig buf %lld -> %lld\n", GST_BUFFER_TIMESTAMP(filter->trig_buf),
		filter->trig_data[filter->trig_n-1]);
}
        filter->in_trigger = TRUE;
	filter->time = trigtime;
        DEBUG("trigger time is %lld\n",filter->time);
        filter->trig_time = filter->time +
		GST_SECOND_MULDIV(filter->chunk_size, filter->rate);
	if(filter->max_rate){
	  guint64 next_time;
	  next_time = filter->time + GST_SECOND_DIV(filter->max_rate);
	  if(filter->trig_time < next_time) filter->trig_time = next_time;
	}
        GST_BUFFER_TIMESTAMP (filter->out_buf) = filter->time;

        offset = ((filter->time - GST_BUFFER_TIMESTAMP (filter->input_buf))
		* filter->rate) / GST_SECOND_DOUBLE;

        do_copy = TRUE;
      }else{
	DEBUG("after endtime\n");
	if(filter->triggered){
          filter->time = filter->trig_data[filter->trig_index] + filter->trig_offset;
	}
      }
    }
  }else{
#if 0
    while (filter->trig_index < filter->trig_n){
      if (filter->time < filter->trig_data[filter->trig_index] + filter->trig_offset) break;
      filter->trig_index++;
    }
#endif
    do_copy = TRUE;
  }

  if(do_copy){
    int out_left;
    int input_left;

    out_left = filter->chunk_size - filter->out_index;
    input_left = filter->input_n - offset;

    if(input_left < out_left){
      DEBUG("copying to end of input len=%d offset=%d\n",input_left,offset);
      memcpy (filter->out_data + filter->out_index,
	  filter->input_data + offset,
	  input_left * sizeof (float));

      filter->out_index += input_left;
      filter->time = filter->input_endtime;
    }else{
      DEBUG("copying to end of output len=%d offset=%d\n",out_left,offset);
      memcpy (filter->out_data + filter->out_index,
	    filter->input_data + offset,
	    out_left * sizeof (float));

      filter->out_index += out_left;
      filter->time = GST_BUFFER_TIMESTAMP(filter->out_buf) +
	      GST_SECOND_MULDIV(filter->chunk_size, filter->rate);
    }
  }

  while (filter->trig_index < filter->trig_n &&
      filter->trig_data[filter->trig_index] + filter->trig_offset < filter->trig_time){
    filter->trig_index++;
  }

  if (filter->out_index >= filter->chunk_size) {
    DEBUG("pushing src\n");
    gst_pad_push (filter->srcpad, filter->out_buf);

    filter->out_buf = NULL;
    filter->in_trigger = FALSE;
    //filter->trig_index++;
  }

  if (filter->trig_index >= filter->trig_n && filter->triggered) {
    DEBUG("unref trig buf\n");
    gst_buffer_unref(filter->trig_buf);
    filter->trig_buf = NULL;
  }

  if (filter->time >= filter->input_endtime) {
    DEBUG("unref input buf\n");
    gst_buffer_unref(filter->input_buf);
    filter->input_buf = NULL;
  }
  DEBUG("return\n");
}

static void
gst_framer_set_property (GObject * object, guint prop_id, const GValue * value,
			 GParamSpec * pspec)
{
  GstFramer *filter;

  g_return_if_fail (GST_IS_FRAMER (object));
  filter = GST_FRAMER (object);

  switch (prop_id) {
    case ARG_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case ARG_CHUNK_SIZE:
      filter->chunk_size = g_value_get_int (value);
      break;
    case ARG_MAX_RATE:
      filter->max_rate = g_value_get_int (value);
      break;
    case ARG_TRIG_OFFSET:
      filter->trig_offset = - g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_framer_get_property (GObject * object, guint prop_id, GValue * value,
			 GParamSpec * pspec)
{
  GstFramer *filter;

  g_return_if_fail (GST_IS_FRAMER (object));
  filter = GST_FRAMER (object);

  switch (prop_id) {
    case ARG_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case ARG_CHUNK_SIZE:
      g_value_set_int (value, filter->chunk_size);
      break;
    case ARG_MAX_RATE:
      g_value_set_int (value, filter->max_rate);
      break;
    case ARG_TRIG_OFFSET:
      g_value_set_int (value, filter->trig_offset);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features 
 */
static gboolean
plugin_init (GModule * module, GstPlugin * plugin)
{
  GstElementFactory *factory;

  factory =
    gst_element_factory_new ("framer", GST_TYPE_FRAMER, &plugin_details);
  g_return_val_if_fail (factory != NULL, FALSE);

  gst_element_factory_add_pad_template (factory, gst_framer_src_factory ());
  gst_element_factory_add_pad_template (factory, gst_framer_sink_factory ());
  gst_element_factory_add_pad_template (factory, gst_framer_sink2_factory ());

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  /* plugin initialisation succeeded */
  return TRUE;
}

/* this is the structure that gst-register looks for
 * so keep the name plugin_desc, or you cannot get your plug-in registered */
GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "framer",
  plugin_init
};
