/* 
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
#include <math.h>
#include <string.h>
#include <gst-sci/gst-sci.h>

#define DEBUG(a...)

/* include this header if you want to use dynamic parameters
#include <gst/control/control.h>
*/

#include <config.h>
#include "gate.h"

static GstElementDetails plugin_details = {
	longname: "Gate",
	klass: "Generic/Gate",
	description: "enable/disable output",
	version: VERSION,
	author: "David A. Schleef <ds@schleef.org>",
	copyright: "(C) 2002"
};

/* Filter signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_SILENT
};

GST_PAD_TEMPLATE_FACTORY (gst_gate_sink_factory,
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  NULL		/* no caps */
);

GST_PAD_TEMPLATE_FACTORY (gst_gate_trigger_sink_factory,
  "trigger_sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  NULL		/* no caps */
);

GST_PAD_TEMPLATE_FACTORY (gst_gate_src_factory,
  "src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  NULL		/* no caps */
);

static void	gst_gate_class_init	(GstGateClass *klass);
static void	gst_gate_init		(GstGate *filter);

static void	gst_gate_set_property	(GObject *object, guint prop_id,
                                                 const GValue *value, 
					         GParamSpec *pspec);
static void	gst_gate_get_property    (GObject *object, guint prop_id,
                                                 GValue *value, 
						 GParamSpec *pspec);

static void	gst_gate_loop		(GstElement *element);

static GstElementClass *parent_class = NULL;

/* this function handles the connection with other plug-ins */
static GstPadConnectReturn
gst_gate_sink_connect (GstPad *pad, GstCaps *caps)
{
  GstGate *gate;
  int rate;
  
  gate = GST_GATE (gst_pad_get_parent (pad));
  g_return_val_if_fail (gate != NULL, GST_PAD_CONNECT_REFUSED);
  g_return_val_if_fail (GST_IS_GATE (gate), 
                        GST_PAD_CONNECT_REFUSED);

  if (!GST_CAPS_IS_FIXED (caps)) {
    return GST_PAD_CONNECT_DELAYED;
  }
  
  g_return_val_if_fail (gate->srcpad != NULL, GST_PAD_CONNECT_DELAYED);

  gst_caps_get_int(caps,"rate",&rate);
  gate->rate = rate;
  DEBUG("got rate %d\n",rate);

  return gst_pad_try_set_caps(gate->srcpad, caps);
}

static GstPadConnectReturn
gst_gate_trigger_sink_connect (GstPad *pad, GstCaps *caps)
{
  GstGate *gate;
  
  gate = GST_GATE (gst_pad_get_parent (pad));
  g_return_val_if_fail (gate != NULL, GST_PAD_CONNECT_REFUSED);
  g_return_val_if_fail (GST_IS_GATE (gate), 
                        GST_PAD_CONNECT_REFUSED);

  if (!GST_CAPS_IS_FIXED (caps)) {
    return GST_PAD_CONNECT_DELAYED;
  }
  
  g_return_val_if_fail (gate->srcpad != NULL, GST_PAD_CONNECT_DELAYED);

  return GST_PAD_CONNECT_OK;
}

GType
gst_gst_gate_get_type (void) 
{
  static GType plugin_type = 0;

  if (!plugin_type) 
  {
    static const GTypeInfo plugin_info = 
    {
      sizeof (GstGateClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_gate_class_init,
      NULL,
      NULL,
      sizeof (GstGate),
      0,
      (GInstanceInitFunc) gst_gate_init,
    };
    plugin_type = g_type_register_static (GST_TYPE_ELEMENT, 
	                                  "GstGate", 
	                                  &plugin_info, 0);
  }
  return plugin_type;
}

/* initialize the plugin's class */
static void
gst_gate_class_init (GstGateClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property (gobject_class, ARG_SILENT,
    g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
                          FALSE, G_PARAM_READWRITE));
  
  gobject_class->set_property = gst_gate_set_property;
  gobject_class->get_property = gst_gate_get_property;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_gate_init (GstGate *filter)
{
  filter->sinkpad = gst_pad_new_from_template (gst_gate_sink_factory (), 
                                               "sink");
  gst_pad_set_connect_function (filter->sinkpad, gst_gate_sink_connect);
  filter->trigger_sinkpad = gst_pad_new_from_template (
	gst_gate_trigger_sink_factory (), "trigger_sink");
  gst_pad_set_connect_function (filter->trigger_sinkpad, gst_gate_trigger_sink_connect);
  filter->srcpad = gst_pad_new_from_template (gst_gate_src_factory (), 
                                              "src");
  //gst_pad_set_connect_function (filter->srcpad, gst_gate_connect);
  
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->trigger_sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_element_set_loop_function (GST_ELEMENT(filter), gst_gate_loop);
  filter->silent = FALSE;

  filter->trig_offset = 0;
}

/* loop function
 * this function does the actual processing
 */

static void
gst_gate_loop (GstElement *element)
{
  GstGate *gate;
  GstEvent *ev = NULL;

  g_return_if_fail (GST_IS_ELEMENT (element));
  
  gate = GST_GATE (element);
  g_return_if_fail (GST_IS_GATE (gate));

  DEBUG("gate loop\n");
  if(!gate->trig_buf){
    guint64 newtime;

    DEBUG("pulling trig buf\n");
    gate->trig_buf = gst_pad_pull(gate->trigger_sinkpad);
    gate->trig_data = (guint64 *)GST_BUFFER_DATA(gate->trig_buf);
    gate->trig_index = 0;
    gate->trig_n = GST_BUFFER_SIZE(gate->trig_buf)/sizeof(guint64);

    if(gate->trig_n > 0){
      newtime = gate->trig_data[0] + gate->trig_offset;
    }else{
      newtime = GST_BUFFER_TIMESTAMP(gate->trig_buf) + gate->trig_offset;
    }
    if(newtime > gate->trig_time) gate->trig_time = newtime;
    gate->time = gate->trig_time;
    DEBUG("gate->trig_time = %lld newtime=%lld trig_n=%d\n",
	gate->trig_time,newtime,gate->trig_n);
  }

  if(!gate->input_buf){
    DEBUG("pulling input buf\n");
    gate->input_buf = gst_pad_pull(gate->sinkpad);

    DEBUG("input ts = %lld\n",GST_BUFFER_TIMESTAMP(gate->input_buf));
    gate->input_n = GST_BUFFER_SIZE(gate->input_buf)/sizeof(float);
    gate->input_endtime = GST_BUFFER_TIMESTAMP(gate->input_buf) +
      (guint64)GST_SECOND_MULDIV(gate->input_n, gate->rate);
    DEBUG("input %lld -> %lld %d %d\n",gate->input_endtime,
      (guint64)GST_SECOND_MULDIV(gate->input_n, gate->rate),
      gate->input_n, gate->rate);
    gate->input_data = (float *)GST_BUFFER_DATA(gate->input_buf);
  }

  while(gate->trig_index < gate->trig_n &&
      gate->trig_data[gate->trig_index] + gate->trig_offset < gate->trig_time){
    gate->trig_index++;
  }

  if(!gate->in_trigger){
    if(gate->trig_index < gate->trig_n){
      guint64 trigtime;

      trigtime = gate->trig_data[gate->trig_index] + gate->trig_offset;
      DEBUG("trigtime = %lld\n",trigtime);
      if(trigtime < gate->input_endtime){
	int offset;

        DEBUG("got trigger\n");
	gate->in_trigger = TRUE;
	gate->time = trigtime;

	gate->stop_time = gate->time + GST_SECOND;
	gate->trig_time = gate->stop_time;
	
	ev = gst_event_new(GST_EVENT_NEW_MEDIA);
	GST_EVENT_TIMESTAMP(ev) = gate->time;

	offset = ((gate->time - GST_BUFFER_TIMESTAMP(gate->input_buf))
		* gate->rate) / GST_SECOND_DOUBLE;
        DEBUG("time=%lld input_buf=%lld rate=%d\n",
	gate->time, GST_BUFFER_TIMESTAMP(gate->input_buf), gate->rate);
	
	gate->out_buf = gst_buffer_create_sub(gate->input_buf, offset,
		GST_BUFFER_SIZE(gate->input_buf) - offset);
	GST_BUFFER_TIMESTAMP(gate->out_buf) = gate->time;

        gate->time = gate->input_endtime;
	gate->trig_index++;
      }else{
	gate->time = trigtime;
	gate->trig_time = gate->stop_time;
      }
    }
  }else{
    gate->out_buf = gate->input_buf;
    gst_buffer_ref(gate->out_buf);

    gate->time = gate->input_endtime;
    gate->trig_time = gate->time;

    if(gate->time > gate->stop_time){
      gate->in_trigger = FALSE;
    }
  }

  if(ev){
    gst_pad_push(gate->srcpad, GST_BUFFER(ev));
  }

  if(gate->out_buf){
    DEBUG("pushing output buf (size=%d)\n",GST_BUFFER_SIZE(gate->out_buf));
    gst_pad_push(gate->srcpad, gate->out_buf);
    gate->out_buf = NULL;
  }

  if(gate->trig_index >= gate->trig_n){
    DEBUG("unreffing trig buf\n");
    gst_buffer_unref(gate->trig_buf);
    gate->trig_buf = NULL;
  }

  if(gate->time >= gate->input_endtime){
    DEBUG("unreffing input buf\n");
    gst_buffer_unref(gate->input_buf);
    gate->input_buf = NULL;
  }
  DEBUG("done\n");
}

static void
gst_gate_set_property (GObject *object, guint prop_id, 
                              const GValue *value, GParamSpec *pspec)
{
  GstGate *filter;

  g_return_if_fail (GST_IS_GATE (object));
  filter = GST_GATE (object);

  switch (prop_id) 
  {
  case ARG_SILENT:
    filter->silent = g_value_get_boolean (value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gst_gate_get_property (GObject *object, guint prop_id, 
                              GValue *value, GParamSpec *pspec)
{
  GstGate *filter;

  g_return_if_fail (GST_IS_GATE (object));
  filter = GST_GATE (object);
  
  switch (prop_id) {
  case ARG_SILENT:
    g_value_set_boolean (value, filter->silent);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features 
 */
static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_element_factory_new ("gate", GST_TYPE_GATE,
                                     &plugin_details);
  g_return_val_if_fail (factory != NULL, FALSE);
  
  gst_element_factory_add_pad_template (factory, 
                                        gst_gate_src_factory ());
  gst_element_factory_add_pad_template (factory, 
                                        gst_gate_sink_factory ());
  gst_element_factory_add_pad_template (factory, 
                                        gst_gate_trigger_sink_factory ());

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  /* plugin initialisation succeeded */
  return TRUE;
}

/* this is the structure that gst-register looks for
 * so keep the name plugin_desc, or you cannot get your plug-in registered */
GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "gate",
  plugin_init
};
