/* 
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 * Copyright (C) 2003      David Schleef <ds@schleef.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>

/* include this header if you want to use dynamic parameters
#include <gst/control/control.h>
*/

#include <config.h>
#include "hold.h"

static GstElementDetails plugin_details = {
	longname: "Hold",
	klass: "Generic/Hold",
	description: "Hold streams until timestamp",
	version: VERSION,
	author: "David Schleef <ds@schleef.org>",
	copyright: "(C) 2003"
};

/* Filter signals and args */
enum {
  SIGNAL_NEW_MEDIA,
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_PLAY_TIME,
  ARG_SILENT
};

GST_PAD_TEMPLATE_FACTORY (gst_hold_sink_factory,
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "hold_src",
    "audio/raw",
      "format",		GST_PROPS_STRING ("int"),
      "law",		GST_PROPS_INT (0),
      "endianness",	GST_PROPS_INT (G_BYTE_ORDER),
      "signed",		GST_PROPS_BOOLEAN (TRUE),
      "width",		GST_PROPS_INT (16),
      "depth",		GST_PROPS_INT (16),
      "rate",		GST_PROPS_INT (44100),
      "channels",	GST_PROPS_INT (1)
  )
);

GST_PAD_TEMPLATE_FACTORY (gst_hold_src_factory,
  "src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "hold_sink",
    "audio/raw",
      "format",		GST_PROPS_STRING ("int"),
      "law",		GST_PROPS_INT (0),
      "endianness",	GST_PROPS_INT (G_BYTE_ORDER),
      "signed",		GST_PROPS_BOOLEAN (TRUE),
      "width",		GST_PROPS_INT (16),
      "depth",		GST_PROPS_INT (16),
      "rate",		GST_PROPS_INT (44100),
      "channels",	GST_PROPS_INT (1)
  )
);

static void	gst_hold_class_init	(GstHoldClass *klass);
static void	gst_hold_init		(GstHold *filter);

static void	gst_hold_set_property	(GObject *object, guint prop_id,
                                                 const GValue *value, 
					         GParamSpec *pspec);
static void	gst_hold_get_property    (GObject *object, guint prop_id,
                                                 GValue *value, 
						 GParamSpec *pspec);

static void	gst_hold_loop		(GstElement *filter);

static GstElementClass *parent_class = NULL;
static guint gst_hold_signals[LAST_SIGNAL] = { 0 };


static GstCaps *gst_hold_getcaps (GstPad *pad, GstCaps *caps)
{
  GstHold *filter;
  GstCaps *ourcaps;
  GstCaps *peercaps;
  GstPad *otherpad;

  GST_DEBUG(0,"gst_hold_getcaps");
  
  filter = GST_HOLD (gst_pad_get_parent (pad));
  g_return_val_if_fail (filter != NULL, NULL);
  g_return_val_if_fail (GST_IS_HOLD (filter), NULL);
  otherpad = (pad == filter->srcpad ? filter->sinkpad : filter->srcpad);
  
  ourcaps = GST_CAPS_NEW(
    "hold_src",
    "audio/raw",
      "format",		GST_PROPS_STRING ("int"),
      "law",		GST_PROPS_INT (0),
      "endianness",	GST_PROPS_INT (G_BYTE_ORDER),
      "signed",		GST_PROPS_BOOLEAN (TRUE),
      "width",		GST_PROPS_INT (16),
      "depth",		GST_PROPS_INT (16),
      "rate",		GST_PROPS_INT (44100),
      "channels",	GST_PROPS_INT (1)
  );

  peercaps = gst_pad_get_allowed_caps (otherpad);

  caps = gst_caps_intersect (ourcaps, peercaps);

  gst_caps_unref(ourcaps);

  return caps;
}

static GstPadLinkReturn
gst_hold_link (GstPad *pad, GstCaps *caps)
{
  GstHold *hold;
  GstPad *otherpad;

  GST_DEBUG(0,"gst_hold_link");
  hold = GST_HOLD(gst_pad_get_parent(pad));

  if (!GST_CAPS_IS_FIXED (caps)) {
    return GST_PAD_LINK_DELAYED;
  }

  otherpad = (pad == hold->srcpad ? hold->sinkpad : hold->srcpad);

  return gst_pad_try_set_caps (otherpad, caps);
}

GType
gst_gst_hold_get_type (void) 
{
  static GType plugin_type = 0;

  if (!plugin_type) 
  {
    static const GTypeInfo plugin_info = 
    {
      sizeof (GstHoldClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_hold_class_init,
      NULL,
      NULL,
      sizeof (GstHold),
      0,
      (GInstanceInitFunc) gst_hold_init,
    };
    plugin_type = g_type_register_static (GST_TYPE_ELEMENT, 
	                                  "GstHold", 
	                                  &plugin_info, 0);
  }
  return plugin_type;
}

/* initialize the plugin's class */
static void
gst_hold_class_init (GstHoldClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property (gobject_class, ARG_SILENT,
    g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
                          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_PLAY_TIME,
    g_param_spec_uint64 ("play-time", "Play Time", "Time to start playing track",
                          G_GINT64_CONSTANT(0), G_MAXUINT64,
			  G_GINT64_CONSTANT(0), G_PARAM_READWRITE));
  
  gst_hold_signals[SIGNAL_NEW_MEDIA] = 
    g_signal_new ("new_media", G_TYPE_FROM_CLASS(klass), G_SIGNAL_RUN_LAST,
	G_STRUCT_OFFSET(GstHoldClass, new_media), NULL, NULL,
	g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
  
  gobject_class->set_property = gst_hold_set_property;
  gobject_class->get_property = gst_hold_get_property;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_hold_init (GstHold *filter)
{
  GST_FLAG_SET(filter, GST_ELEMENT_EVENT_AWARE);

  filter->sinkpad = gst_pad_new_from_template (gst_hold_sink_factory (), 
                                               "sink");
  gst_pad_set_getcaps_function (filter->sinkpad, gst_hold_getcaps);
  gst_pad_set_link_function (filter->sinkpad, gst_hold_link);

  filter->srcpad = gst_pad_new_from_template (gst_hold_src_factory (), 
                                              "src");
  gst_pad_set_getcaps_function (filter->srcpad, gst_hold_getcaps);
  gst_pad_set_link_function (filter->srcpad, gst_hold_link);
  
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;

  gst_element_set_loop_function (GST_ELEMENT(filter), (void *)gst_hold_loop);

  filter->playing = FALSE;
  filter->play_time = GST_CLOCK_TIME_NONE;
  filter->chunk_size = 2048;
  filter->rate = 44100;
}

static void
gst_hold_loop (GstElement *element)
{
  GstBuffer *buffer;
  GstHold *filter;

  g_return_if_fail (GST_IS_HOLD (element));
  filter = GST_HOLD(element);

  if (filter->playing == TRUE){
    //g_print("playing: %" G_GUINT64_FORMAT "\n",(filter->n_samples * GST_SECOND) / filter->rate);

    buffer = gst_pad_pull (filter->sinkpad);
    GST_BUFFER_TIMESTAMP(buffer) = (filter->n_samples * GST_SECOND) / filter->rate;

    filter->n_samples += GST_BUFFER_SIZE(buffer)/2;

    if(GST_IS_EVENT(buffer)){
      GstEvent *event = GST_EVENT (buffer);

      switch (GST_EVENT_TYPE (event)) {
	case GST_EVENT_EOS:
	  filter->playing = FALSE;
          //g_print("stop playing: %" G_GUINT64_FORMAT "\n", (filter->n_samples * GST_SECOND) / filter->rate);
          g_signal_emit(G_OBJECT(filter), gst_hold_signals[SIGNAL_NEW_MEDIA],
              0, NULL);
	  break;
	default:
	  gst_pad_event_default (filter->sinkpad, event);
	  break;
      }
    }else{
      //g_print("pushing %" G_GUINT64_FORMAT "\n",GST_BUFFER_TIMESTAMP(buffer));
      gst_pad_push (filter->srcpad, buffer);
    }
  }else{
    void *data;
    int size;
    GstClockTime next_timestamp;

    //g_print("silent: %" G_GUINT64_FORMAT " %" G_GUINT64_FORMAT "\n", (filter->n_samples * GST_SECOND) / filter->rate, filter->play_time);

    buffer = gst_buffer_new();

    /* FIXME: This isn't sample-accurate */
    size = filter->chunk_size*2;
    data = g_malloc0(size);
    //((short *)data)[0] = 0x0fff;
    GST_BUFFER_DATA(buffer) = data;
    GST_BUFFER_SIZE(buffer) = size;
    GST_BUFFER_TIMESTAMP(buffer) = (filter->n_samples * GST_SECOND) / filter->rate;

    filter->n_samples += filter->chunk_size;

    //g_print("pushing %" G_GUINT64_FORMAT "\n",GST_BUFFER_TIMESTAMP(buffer));
    gst_pad_push (filter->srcpad, buffer);

    next_timestamp = (filter->n_samples * GST_SECOND) / filter->rate;
    if(filter->play_time != GST_CLOCK_TIME_NONE && 
	next_timestamp >= filter->play_time){

      //g_print("start playing: %" G_GUINT64_FORMAT "\n", (filter->n_samples * GST_SECOND) / filter->rate);

      filter->playing = TRUE;
      filter->play_time = GST_CLOCK_TIME_NONE;
    }
  }
}

static void
gst_hold_set_property (GObject *object, guint prop_id, 
                              const GValue *value, GParamSpec *pspec)
{
  GstHold *filter;

  g_return_if_fail (GST_IS_HOLD (object));
  filter = GST_HOLD (object);

  switch (prop_id) 
  {
  case ARG_SILENT:
    filter->silent = g_value_get_boolean (value);
    break;
  case ARG_PLAY_TIME:
    //g_print("set play_time\n");
    filter->play_time = g_value_get_uint64 (value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gst_hold_get_property (GObject *object, guint prop_id, 
                              GValue *value, GParamSpec *pspec)
{
  GstHold *filter;

  g_return_if_fail (GST_IS_HOLD (object));
  filter = GST_HOLD (object);
  
  switch (prop_id) {
  case ARG_SILENT:
    g_value_set_boolean (value, filter->silent);
    break;
  case ARG_PLAY_TIME:
    g_value_set_uint64 (value, filter->play_time);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features 
 */
static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_element_factory_new ("hold", GST_TYPE_HOLD,
                                     &plugin_details);
  g_return_val_if_fail (factory != NULL, FALSE);
  
  gst_element_factory_add_pad_template (factory, 
                                        gst_hold_src_factory ());
  gst_element_factory_add_pad_template (factory, 
                                        gst_hold_sink_factory ());

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  /* plugin initialisation succeeded */
  return TRUE;
}

/* this is the structure that gst-register looks for
 * so keep the name plugin_desc, or you cannot get your plug-in registered */
GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "hold",
  plugin_init
};
