/* 
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 * Copyright (C) 2002      David A. Schleef <ds@schleef.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
#include <math.h>
#include <string.h>

/* include this header if you want to use dynamic parameters
#include <gst/control/control.h>
*/

#include <config.h>
#include "sonogram.h"

static GstElementDetails plugin_details = {
	longname: "Sonogram",
	klass: "Generic/Sonogram",
//[license] = "LGPL",
	description: "Creates sonogram video streams",
	version: VERSION,
	author: "David Schleef <ds@schleef.org>",
	copyright: "(C) 2002"
};

/* Filter signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_SILENT,
  ARG_GAMMA,
  ARG_WIDTH,
  ARG_HEIGHT,
};

GST_PAD_TEMPLATE_FACTORY (gst_sonogram_sink_factory,
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "sonogram_sink",
    "x-gst-sci/frame",
      "format",		GST_PROPS_STRING("float32"),
      "rate",		GST_PROPS_INT_RANGE(1, 1000000000),
      "channels",	GST_PROPS_INT(1)
  )
);

GST_PAD_TEMPLATE_FACTORY (gst_sonogram_src_factory,
  "src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "sonogram_src",
    "video/raw",
      "format", GST_PROPS_FOURCC(GST_MAKE_FOURCC('I','4','2','0')),
      "width", GST_PROPS_INT_RANGE(0,G_MAXINT),
      "height", GST_PROPS_INT_RANGE(0,G_MAXINT)
  )
);

static void	gst_sonogram_class_init	(GstSonogramClass *klass);
static void	gst_sonogram_init		(GstSonogram *filter);

static void	gst_sonogram_set_property	(GObject *object, guint prop_id,
                                                 const GValue *value, 
					         GParamSpec *pspec);
static void	gst_sonogram_get_property    (GObject *object, guint prop_id,
                                                 GValue *value, 
						 GParamSpec *pspec);

static void gst_sonogram_reinit(GstSonogram *filter);
static void	gst_sonogram_chain		(GstPad *pad, GstBuffer *buf);

static GstElementClass *parent_class = NULL;

/* this function handles the connection with other plug-ins */
static GstPadConnectReturn
gst_sonogram_connect (GstPad *pad, GstCaps *caps)
{
  GstSonogram *filter;
  GstPad *otherpad;
  
  filter = GST_SONOGRAM (gst_pad_get_parent (pad));
  g_return_val_if_fail (filter != NULL, GST_PAD_CONNECT_REFUSED);
  g_return_val_if_fail (GST_IS_SONOGRAM (filter), 
                        GST_PAD_CONNECT_REFUSED);
  otherpad = (pad == filter->srcpad ? filter->sinkpad : filter->srcpad);
  
  if (!GST_CAPS_IS_FIXED (caps)) {
	  return GST_PAD_CONNECT_DELAYED;
  }

  return gst_pad_try_set_caps (filter->srcpad,
  	GST_CAPS_NEW(
    "sonogram_src",
    "video/raw",
      "format", GST_PROPS_FOURCC(GST_MAKE_FOURCC('I','4','2','0')),
      "width", GST_PROPS_INT(filter->width),
      "height", GST_PROPS_INT(filter->height)
  ));
}

GType
gst_gst_sonogram_get_type (void) 
{
  static GType plugin_type = 0;

  if (!plugin_type) 
  {
    static const GTypeInfo plugin_info = 
    {
      sizeof (GstSonogramClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_sonogram_class_init,
      NULL,
      NULL,
      sizeof (GstSonogram),
      0,
      (GInstanceInitFunc) gst_sonogram_init,
    };
    plugin_type = g_type_register_static (GST_TYPE_ELEMENT, 
	                                  "GstSonogram", 
	                                  &plugin_info, 0);
  }
  return plugin_type;
}

/* initialize the plugin's class */
static void
gst_sonogram_class_init (GstSonogramClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property (gobject_class, ARG_SILENT,
    g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
                          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_GAMMA,
    g_param_spec_float ("gamma", "Gamma", "Gamma to apply to input",
            0.0, 1.0, 0.25, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_WIDTH,
    g_param_spec_int ("width", "Width", "Width of video stream",
            1, G_MAXINT, 320, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_HEIGHT,
    g_param_spec_int ("height", "Height", "Height of video stream",
            1, G_MAXINT, 240, G_PARAM_READWRITE));
  
  gobject_class->set_property = gst_sonogram_set_property;
  gobject_class->get_property = gst_sonogram_get_property;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_sonogram_init (GstSonogram *filter)
{
  filter->sinkpad = gst_pad_new_from_template (gst_sonogram_sink_factory (), 
                                               "sink");
  gst_pad_set_connect_function (filter->sinkpad, gst_sonogram_connect);
  filter->srcpad = gst_pad_new_from_template (gst_sonogram_src_factory (), 
                                              "src");
  gst_pad_set_connect_function (filter->srcpad, gst_sonogram_connect);
  
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_pad_set_chain_function (filter->sinkpad, gst_sonogram_chain);
  filter->silent = FALSE;
  filter->gamma = 0.25;
  filter->width = 320;
  filter->height = 240;

  gst_sonogram_reinit(filter);
}

/* chain function
 * this function does the actual processing
 */

static void gst_sonogram_reinit(GstSonogram *filter)
{
	int out_len = (filter->width*filter->height*3) >>1;
	int y_len = filter->width*filter->height;

	if(filter->out_data){
		free(filter->out_data);
	}

	filter->out_len = out_len;
	filter->out_data = g_malloc(filter->out_len);
	memset(filter->out_data, 0, y_len);
	memset(filter->out_data + y_len, 128, out_len - y_len);

	if(!filter->outbuf){
		filter->outbuf = gst_buffer_new();
		gst_buffer_ref(filter->outbuf);
	}
	GST_BUFFER_DATA(filter->outbuf) = (void *)filter->out_data;
	GST_BUFFER_SIZE(filter->outbuf) = filter->out_len;
}

#if 0
static void
extents_f32(float *p_min, float *p_max, float *src, int n)
{
	float min, max;
	int i;

	max = min = src[0];
	for(i=1;i<n;i++){
		if(min>src[i])min = src[i];
		if(max<src[i])max = src[i];
	}
	*p_min = min;
	*p_max = max;
}
#endif

static void
extents_f64(double *p_min, double *p_max, double *src, int n)
{
	double min, max;
	int i;

	max = min = src[0];
	for(i=1;i<n;i++){
		if(min>src[i])min = src[i];
		if(max<src[i])max = src[i];
	}
	*p_min = min;
	*p_max = max;
}

static void
rotate_left_I420(unsigned char *data, int width, int height, int n)
{
	int i;

	for(i=0;i<height;i++){
		memmove(data, data + n, width-n);
		data += width;
	}
}

static void
gst_sonogram_chain (GstPad *pad, GstBuffer *buf)
{
  GstSonogram *filter;
  GstBuffer *outbuf;
  float *data;
  double min, max;
  double mult;
  double x;
  int i;
  double *ldata;
  int in_len;

  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);
  
  filter = GST_SONOGRAM (GST_OBJECT_PARENT (pad));
  g_return_if_fail (GST_IS_SONOGRAM (filter));

  outbuf = gst_buffer_copy(filter->outbuf);
  GST_BUFFER_TIMESTAMP(outbuf) = GST_BUFFER_TIMESTAMP(buf);
  
  rotate_left_I420(filter->out_data, filter->width, filter->height, 1);

  data = (float *)GST_BUFFER_DATA(buf);
  in_len = GST_BUFFER_SIZE(buf)/sizeof(float);
  ldata = malloc(sizeof(double)*filter->height);
  for(i=0;i<filter->height && i<in_len;i++){
	//ldata[i] = log10(data[i]);
	//ldata[i] = data[i];
	ldata[i] = pow(data[i],filter->gamma);
  }
  for(;i<filter->height;i++){
	ldata[i] = 0;
  }
  extents_f64(&min, &max, ldata, filter->height);
#if 1
  if(min<0)min=0;
  if(max<0)max=0;

  if(min<filter->min){
	  filter->min = min;
  }else{
	  filter->min = 0.99 * filter->min + 0.01 * min;
  }

  if(max>filter->max){
	  filter->max = max;
  }else{
	  filter->max = 0.99 * filter->max + 0.01 * max;
  }

  printf("min=%g max=%g\n", filter->min, filter->max);
#else
  filter->max = max;
  filter->min = min;
  printf("min=%g max=%g\n", filter->min, filter->max);
#endif

  mult = 223.0/(filter->max - filter->min);
  for(i=0;i<filter->height;i++){
	x = (ldata[i] - filter->min)*mult;
	if(x<0)x=0;
	if(x>223)x=223;
	filter->out_data[filter->width*(filter->height - i - 1) - 1] = 16 + floor(x);
  }

  free(ldata);

  gst_buffer_unref(buf);

  gst_pad_push (filter->srcpad, outbuf);
}

static void
gst_sonogram_set_property (GObject *object, guint prop_id, 
                              const GValue *value, GParamSpec *pspec)
{
  GstSonogram *filter;

  g_return_if_fail (GST_IS_SONOGRAM (object));
  filter = GST_SONOGRAM (object);

  switch (prop_id) 
  {
  case ARG_SILENT:
    filter->silent = g_value_get_boolean (value);
    break;
  case ARG_GAMMA:
    filter->gamma = g_value_get_float (value);
    break;
  case ARG_WIDTH:
    filter->width = g_value_get_int (value);
    gst_sonogram_reinit(filter);
    break;
  case ARG_HEIGHT:
    filter->height = g_value_get_int (value);
    gst_sonogram_reinit(filter);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gst_sonogram_get_property (GObject *object, guint prop_id, 
                              GValue *value, GParamSpec *pspec)
{
  GstSonogram *filter;

  g_return_if_fail (GST_IS_SONOGRAM (object));
  filter = GST_SONOGRAM (object);
  
  switch (prop_id) {
  case ARG_SILENT:
    g_value_set_boolean (value, filter->silent);
    break;
  case ARG_GAMMA:
    g_value_set_float (value, filter->gamma);
    break;
  case ARG_WIDTH:
    g_value_set_float (value, filter->width);
    break;
  case ARG_HEIGHT:
    g_value_set_float (value, filter->height);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features 
 */
static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_element_factory_new ("sonogram", GST_TYPE_SONOGRAM,
                                     &plugin_details);
  g_return_val_if_fail (factory != NULL, FALSE);
  
  gst_element_factory_add_pad_template (factory, 
                                        gst_sonogram_src_factory ());
  gst_element_factory_add_pad_template (factory, 
                                        gst_sonogram_sink_factory ());

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  /* plugin initialisation succeeded */
  return TRUE;
}

/* this is the structure that gst-register looks for
 * so keep the name plugin_desc, or you cannot get your plug-in registered */
GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "sonogram",
  plugin_init
};
