/* 
 * 
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
#include <math.h>
#include <stdlib.h>

/* include this header if you want to use dynamic parameters
#include <gst/control/control.h>
*/

#include <config.h>
#include "testsrc.h"

static GstElementDetails plugin_details = {
  "Testsrc",
  "Generic/Testsrc",
  "Data Test Source",
  VERSION,
  "David Schleef <ds@schleef.org>",
  "(C) 2002"
};

/* Filter signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_SILENT,
  ARG_RATE,
  ARG_CHUNKSIZE,
  ARG_FREQUENCY,
  ARG_AMPLITUDE,
  ARG_SYNC,
  ARG_NOISE,
  ARG_FREQUENCY_NOISE,
};

GST_PAD_TEMPLATE_FACTORY (gst_testsrc_src_factory,
  "src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "testsrc_src",
    "x-gst-sci/time-series",
      "format",		GST_PROPS_STRING("float32"),
      "axis1",		GST_PROPS_STRING("test input (unitless)"),
      "rate",		GST_PROPS_INT_RANGE(1, 1000000000),
      "channels",	GST_PROPS_INT_RANGE(1, 16)
  )
);

static void	gst_testsrc_class_init	(GstTestsrcClass *klass);
static void	gst_testsrc_init		(GstTestsrc *filter);

static GstBuffer * gst_testsrc_get (GstPad *pad);
static void	gst_testsrc_set_property	(GObject *object, guint prop_id,
                                                 const GValue *value, 
					         GParamSpec *pspec);
static void	gst_testsrc_get_property    (GObject *object, guint prop_id,
                                                 GValue *value, 
						 GParamSpec *pspec);

static GstElementClass *parent_class = NULL;

/* this function handles the connection with other plug-ins */
static GstPadConnectReturn
gst_testsrc_connect (GstPad *pad, GstCaps *caps)
{
  GstTestsrc *filter;
  
  filter = GST_TESTSRC (gst_pad_get_parent (pad));
  g_return_val_if_fail (filter != NULL, GST_PAD_CONNECT_REFUSED);
  g_return_val_if_fail (GST_IS_TESTSRC (filter), 
                        GST_PAD_CONNECT_REFUSED);
  
  if (GST_CAPS_IS_FIXED (caps)) 
  {
    /* caps are not fixed, so try to connect on the other side and see if
     * that works */

    /* caps on other side were accepted, so we're ok */
    return GST_PAD_CONNECT_OK;
  }
  /* not enough information yet, delay negotation */ 
  return GST_PAD_CONNECT_DELAYED;
}

GType
gst_gst_testsrc_get_type (void) 
{
  static GType plugin_type = 0;

  if (!plugin_type) 
  {
    static const GTypeInfo plugin_info = 
    {
      sizeof (GstTestsrcClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_testsrc_class_init,
      NULL,
      NULL,
      sizeof (GstTestsrc),
      0,
      (GInstanceInitFunc) gst_testsrc_init,
    };
    plugin_type = g_type_register_static (GST_TYPE_ELEMENT, 
	                                  "GstTestsrc", 
	                                  &plugin_info, 0);
  }
  return plugin_type;
}

/* initialize the plugin's class */
static void
gst_testsrc_class_init (GstTestsrcClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property (gobject_class, ARG_SILENT,
    g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
                          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_RATE,
    g_param_spec_int ("rate", "Rate", "Output rate",
	    1, 1000000000, 10000, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_CHUNKSIZE,
    g_param_spec_int ("chunk-size", "Chunk size", "Samples per chunk",
	    1, 65536, 1024, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_FREQUENCY,
    g_param_spec_float ("frequency", "Frequency", "Output frequency",
	    0.0, G_MAXFLOAT, 1000.0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_AMPLITUDE,
    g_param_spec_float ("amplitude", "Amplitude", "Output amplitude",
	    0.0, G_MAXFLOAT, 1.0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_SYNC,
    g_param_spec_boolean ("sync", "Sync", "Synchronize output",
                          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_NOISE,
    g_param_spec_float ("noise", "Noise", "Noise level",
	    0.0, G_MAXFLOAT, 0.0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_FREQUENCY_NOISE,
    g_param_spec_float ("frequency_noise", "Frequency noise", "Frequency noise",
	    0.0, G_MAXFLOAT, 0.0, G_PARAM_READWRITE));
  
  gobject_class->set_property = gst_testsrc_set_property;
  gobject_class->get_property = gst_testsrc_get_property;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_testsrc_init (GstTestsrc *filter)
{
  filter->srcpad = gst_pad_new_from_template (gst_testsrc_src_factory (), 
                                              "src");
  gst_pad_set_connect_function (filter->srcpad, gst_testsrc_connect);
  gst_pad_set_get_function(filter->srcpad, gst_testsrc_get);
  
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;
  filter->sync = TRUE;
  filter->channels = 1;
  filter->chunk_size = 1024;
  filter->rate = 10000;
  filter->frequency = 1000;
  filter->frequency_noise = 0.0;
  filter->noise = 0.0;
  filter->amplitude = 1.0;
}

/* chain function
 * this function does the actual processing
 */

#define frand() (rand()/(RAND_MAX+1.0))
static double random_gaussian(void)
{
	double x,y;

	do{
		x = 6*(frand()-0.5);
		y = frand();
	}while(exp(-x*x)<y);

	return x;
}

static GstBuffer *
gst_testsrc_get (GstPad *pad)
{
  GstTestsrc *src;
  GstBuffer *buf;
  float *data;
  int i;
  double acc_increment;
  double acc_increment_noise;

  g_return_val_if_fail (GST_IS_PAD (pad), NULL);
  
  src = GST_TESTSRC (GST_OBJECT_PARENT (pad));
  g_return_val_if_fail (GST_IS_TESTSRC (src), NULL);

  if(!GST_PAD_CAPS(pad)){
	  if(!gst_pad_try_set_caps(src->srcpad,
		GST_CAPS_NEW(
    "testsrc_src",
    "x-gst-sci/time-series",
      "format",		GST_PROPS_STRING("float32"),
      "axis1",		GST_PROPS_STRING("test input (unitless)"),
      "rate",		GST_PROPS_INT(src->rate),
      "channels",	GST_PROPS_INT(src->channels)
		))){
		gst_element_error(GST_ELEMENT(src), "could not set caps");

		return NULL;
	}
  }

  buf = gst_buffer_new();
  g_return_val_if_fail(buf, NULL);

  data = g_malloc(src->chunk_size * sizeof(float));
  GST_BUFFER_DATA(buf) = (gpointer)data;
  GST_BUFFER_SIZE(buf) = src->chunk_size * sizeof(float);
  //GST_BUFFER_OFFSET(buf) = src->curoffset;
  GST_BUFFER_TIMESTAMP(buf) = src->timestamp;
  src->timestamp += GST_SECOND*src->chunk_size/src->rate;

  if(src->sync){
	double chunk_time_us;
  	chunk_time_us = 1000000.0*src->chunk_size/src->rate;
  	usleep(chunk_time_us);
  }

  acc_increment = 2*M_PI*src->frequency/src->rate;
  acc_increment_noise = 2*M_PI*src->frequency_noise/src->rate;
  for(i=0;i<src->chunk_size;i++){
	  data[i] = src->amplitude * sin(src->acc);
	  data[i] += src->noise * random_gaussian();

	  src->acc += acc_increment;
	  src->acc += acc_increment_noise * random_gaussian();
	  if(src->acc>=2*M_PI)src->acc-=2*M_PI;
  }

  src->curoffset += src->chunk_size * sizeof(float);

  return buf;
}

static void
gst_testsrc_set_property (GObject *object, guint prop_id, 
                              const GValue *value, GParamSpec *pspec)
{
  GstTestsrc *filter;

  g_return_if_fail (GST_IS_TESTSRC (object));
  filter = GST_TESTSRC (object);

  switch (prop_id) 
  {
  case ARG_SILENT:
    filter->silent = g_value_get_boolean (value);
    break;
  case ARG_RATE:
    filter->rate = g_value_get_int (value);
    break;
  case ARG_CHUNKSIZE:
    filter->chunk_size = g_value_get_int (value);
    break;
  case ARG_FREQUENCY:
    filter->frequency = g_value_get_float (value);
    break;
  case ARG_AMPLITUDE:
    filter->amplitude = g_value_get_float (value);
    break;
  case ARG_SYNC:
    filter->sync = g_value_get_boolean (value);
    break;
  case ARG_NOISE:
    filter->noise = g_value_get_float (value);
    break;
  case ARG_FREQUENCY_NOISE:
    filter->frequency_noise = g_value_get_float (value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gst_testsrc_get_property (GObject *object, guint prop_id, 
                              GValue *value, GParamSpec *pspec)
{
  GstTestsrc *filter;

  g_return_if_fail (GST_IS_TESTSRC (object));
  filter = GST_TESTSRC (object);
  
  switch (prop_id) {
  case ARG_SILENT:
    g_value_set_boolean (value, filter->silent);
    break;
  case ARG_RATE:
    g_value_set_int (value, filter->rate);
    break;
  case ARG_CHUNKSIZE:
    g_value_set_int (value, filter->chunk_size);
    break;
  case ARG_FREQUENCY:
    g_value_set_float (value, filter->frequency);
    break;
  case ARG_AMPLITUDE:
    g_value_set_float (value, filter->amplitude);
    break;
  case ARG_SYNC:
    g_value_set_boolean (value, filter->silent);
    break;
  case ARG_NOISE:
    g_value_set_float (value, filter->noise);
    break;
  case ARG_FREQUENCY_NOISE:
    g_value_set_float (value, filter->frequency_noise);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features 
 */
static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_element_factory_new ("testsrc", GST_TYPE_TESTSRC,
                                     &plugin_details);
  g_return_val_if_fail (factory != NULL, FALSE);
  
  gst_element_factory_add_pad_template (factory, 
                                        gst_testsrc_src_factory ());

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  /* plugin initialisation succeeded */
  return TRUE;
}

/* this is the structure that gst-register looks for
 * so keep the name plugin_desc, or you cannot get your plug-in registered */
GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "testsrc",
  plugin_init
};
