/* vi: set ts=2: */
/* 
 * GStreamer
 * Copyright (C) 1999-2001 Erik Walthinsen <omega@cse.ogi.edu>
 * Copyright (C) 2002      David A. Schleef <ds@schleef.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>

/* include this header if you want to use dynamic parameters
#include <gst/control/control.h>
*/

#include <config.h>
#include "trigger.h"

static GstElementDetails plugin_details = {
	longname:	"Trigger",
	klass:		"Filter/Trigger",
	description:	"Create timestamp stream from input",
	version:	VERSION,
	author:		"David A. Schleef <ds@schleef.org>",
	copyright:	"(C) 2002"
};

/* Filter signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_SILENT,
	ARG_SET_LEVEL,
	ARG_RESET_LEVEL,
};

GST_PAD_TEMPLATE_FACTORY (gst_trigger_sink_factory,
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "trigger_sink",
    "x-gst-sci/time-series",
      "format",		GST_PROPS_STRING("float32"),
      "rate",		GST_PROPS_INT_RANGE(1, 1000000000),
      "channels",	GST_PROPS_INT(1)
  )
);

GST_PAD_TEMPLATE_FACTORY (gst_trigger_src_factory,
  "src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW(
    "trigger_src",
    "x-gst-sci/timestamp",
    NULL
  )
);

static void	gst_trigger_class_init	(GstTriggerClass *klass);
static void	gst_trigger_init		(GstTrigger *filter);

static void	gst_trigger_set_property	(GObject *object, guint prop_id,
                                                 const GValue *value, 
					         GParamSpec *pspec);
static void	gst_trigger_get_property    (GObject *object, guint prop_id,
                                                 GValue *value, 
						 GParamSpec *pspec);

static void	gst_trigger_chain		(GstPad *pad, GstBuffer *buf);

static GstElementClass *parent_class = NULL;

/* this function handles the connection with other plug-ins */
static GstPadConnectReturn
gst_trigger_connect (GstPad *pad, GstCaps *caps)
{
  GstTrigger *filter;
  
  filter = GST_TRIGGER (gst_pad_get_parent (pad));
  g_return_val_if_fail (filter != NULL, GST_PAD_CONNECT_REFUSED);
  g_return_val_if_fail (GST_IS_TRIGGER (filter), 
                        GST_PAD_CONNECT_REFUSED);
  
  if (!GST_CAPS_IS_FIXED (caps)) 
  	return GST_PAD_CONNECT_DELAYED;

	if(!gst_caps_has_property(caps,"rate"))
		return GST_PAD_CONNECT_DELAYED;

  gst_caps_get_int(caps,"rate",&filter->rate);
	//filter->rate = 10000;

  return gst_pad_try_set_caps(filter->srcpad,
	GST_CAPS_NEW(
    "trigger_src",
    "x-gst-sci/timestamp",
    NULL
  ));
}

GType
gst_gst_trigger_get_type (void) 
{
  static GType plugin_type = 0;

  if (!plugin_type) 
  {
    static const GTypeInfo plugin_info = 
    {
      sizeof (GstTriggerClass),
      NULL,
      NULL,
      (GClassInitFunc) gst_trigger_class_init,
      NULL,
      NULL,
      sizeof (GstTrigger),
      0,
      (GInstanceInitFunc) gst_trigger_init,
    };
    plugin_type = g_type_register_static (GST_TYPE_ELEMENT, 
	                                  "GstTrigger", 
	                                  &plugin_info, 0);
  }
  return plugin_type;
}

/* initialize the plugin's class */
static void
gst_trigger_class_init (GstTriggerClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_ref (GST_TYPE_ELEMENT);

  g_object_class_install_property (gobject_class, ARG_SILENT,
    g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
                          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_SET_LEVEL,
    g_param_spec_float ("set_level", "Set Level", "Set Level",
						-G_MAXFLOAT, G_MAXFLOAT, 0.0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_RESET_LEVEL,
    g_param_spec_float ("reset_level", "Reset Level", "Reset Level",
						-G_MAXFLOAT, G_MAXFLOAT, 0.0, G_PARAM_READWRITE));
  
  gobject_class->set_property = gst_trigger_set_property;
  gobject_class->get_property = gst_trigger_get_property;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_trigger_init (GstTrigger *filter)
{
  filter->sinkpad = gst_pad_new_from_template (gst_trigger_sink_factory (), 
                                               "sink");
  gst_pad_set_connect_function (filter->sinkpad, gst_trigger_connect);
  filter->srcpad = gst_pad_new_from_template (gst_trigger_src_factory (), 
                                              "src");
  gst_pad_set_connect_function (filter->srcpad, gst_trigger_connect);
  
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_pad_set_chain_function (filter->sinkpad, gst_trigger_chain);
  filter->silent = FALSE;

}

/* chain function
 * this function does the actual processing
 */

#define TRIG_SET 1
#define TRIG_RESET 0

static void
gst_trigger_chain (GstPad *pad, GstBuffer *buf)
{
  GstTrigger *filter;
  GstBuffer *newbuf;
  guint64 *data;
  int n_alloc;
  int n;
  float *i_data;
  int buflen;
  int state;
  int i;

  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);
  
  filter = GST_TRIGGER (GST_OBJECT_PARENT (pad));
  g_return_if_fail (GST_IS_TRIGGER (filter));

  buflen = GST_BUFFER_SIZE(buf)/sizeof(float);
  i_data = (float *)GST_BUFFER_DATA(buf);

  newbuf = gst_buffer_new();
  
  n_alloc = 16;
  data = g_malloc(n_alloc*sizeof(guint64));
  n = 0;
  state = filter->state;
  for(i=0;i<buflen;i++){
	  int oldstate = state;
	  if(state == TRIG_SET){
		  if(i_data[i]<filter->reset_level){
			  state = TRIG_RESET;
		  }
	  }else{
		  if(i_data[i]>filter->set_level){
			  state = TRIG_SET;
		  }
	  }
	  if(state!=oldstate && state==TRIG_SET){
		  if(n>=n_alloc){
			  n_alloc <<= 1;
			  data = g_realloc(data,n_alloc*sizeof(guint64));
		  }
		  data[n] = GST_BUFFER_TIMESTAMP(buf) + (i*((double)GST_SECOND))/filter->rate;
			n++;
	  }
  }
  filter->state = state;

  GST_BUFFER_SIZE(newbuf) = n*sizeof(guint64);
  GST_BUFFER_DATA(newbuf) = (void *)data;
  GST_BUFFER_TIMESTAMP(newbuf) = GST_BUFFER_TIMESTAMP(buf);

  gst_buffer_unref(buf);
  gst_pad_push (filter->srcpad, newbuf);
}

static void
gst_trigger_set_property (GObject *object, guint prop_id, 
                              const GValue *value, GParamSpec *pspec)
{
  GstTrigger *filter;

  g_return_if_fail (GST_IS_TRIGGER (object));
  filter = GST_TRIGGER (object);

  switch (prop_id) 
  {
  case ARG_SILENT:
    filter->silent = g_value_get_boolean (value);
    break;
	case ARG_SET_LEVEL:
    filter->set_level = g_value_get_float (value);
    break;
	case ARG_RESET_LEVEL:
    filter->reset_level = g_value_get_float (value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gst_trigger_get_property (GObject *object, guint prop_id, 
                              GValue *value, GParamSpec *pspec)
{
  GstTrigger *filter;

  g_return_if_fail (GST_IS_TRIGGER (object));
  filter = GST_TRIGGER (object);
  
  switch (prop_id) {
  case ARG_SILENT:
    g_value_set_boolean (value, filter->silent);
    break;
	case ARG_SET_LEVEL:
    g_value_set_float (value, filter->set_level);
    break;
	case ARG_RESET_LEVEL:
    g_value_set_float (value, filter->reset_level);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features 
 */
static gboolean
plugin_init (GModule *module, GstPlugin *plugin)
{
  GstElementFactory *factory;

  factory = gst_element_factory_new ("trigger", GST_TYPE_TRIGGER,
                                     &plugin_details);
  g_return_val_if_fail (factory != NULL, FALSE);
  
  gst_element_factory_add_pad_template (factory, 
                                        gst_trigger_src_factory ());
  gst_element_factory_add_pad_template (factory, 
                                        gst_trigger_sink_factory ());

  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));

  /* plugin initialisation succeeded */
  return TRUE;
}

/* this is the structure that gst-register looks for
 * so keep the name plugin_desc, or you cannot get your plug-in registered */
GstPluginDesc plugin_desc = {
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "trigger",
  plugin_init
};
